# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Layout representation
~~~~~~~~~~~~~~~~~~~~~
Classes for representing layouts of data
"""
import dataclasses
import warnings
from functools import cached_property, singledispatch
from math import ceil
from typing import Any, Literal, Mapping, Optional, Sequence, Tuple, Union

BIG = 2**63 - 1


@dataclasses.dataclass(frozen=True)
class CellLayoutArgument:
    # We use this as a buffer for operators with precedence slower than <<,
    # i.e. | (__or__)
    layout: 'CellLayout'
    argument: Any


@dataclasses.dataclass(frozen=True)
class CellLayoutSubstitution:
    layout: 'CellLayout'
    substitute: 'CellLayout'

    def __lshift__(self, other: int):
        return substitute(self.layout, self.substitute, other)


@dataclasses.dataclass(frozen=True)
class CellLayoutFloat:
    anchor: 'CellLayout'
    floating: 'CellLayout'

    def __lshift__(
        self,
        other: Tuple[Tuple[float, float], Tuple[float, float]],
    ):
        return float_layout(
            self.floating,
            self.anchor,
            loc_rel=other[0],
            dim_rel=other[1],
        )


@dataclasses.dataclass(frozen=True)
class CellLayoutVerticalChain:
    chain: Sequence['CellLayout']

    def __or__(self, other: Union[float, 'CellLayoutArgument']):
        if isinstance(other, CellLayoutArgument):
            # We already have everything required to evaluate the chain
            chain = self.chain + [other.layout]
            return vsplit(other.argument, *chain)
        return CellLayoutVerticalChain(
            chain=self.chain + [other]
        )

    def __lshift__(self, other: float):
        return vsplit(other, *self.chain)


@dataclasses.dataclass(frozen=True)
class CellLayoutHorizontalChain:
    chain: Sequence['CellLayout']

    def __truediv__(self, other: 'CellLayout'):
        return CellLayoutHorizontalChain(
            chain=self.chain + [other]
        )

    def __lshift__(self, other: float):
        return hsplit(other, *self.chain)


class CellLayout:
    """
    Binary tree layout of cells, addressable and iterable in hierarchical
    left-to-right and top-to-bottom order
    """
    # It's mutable and TBH I hate it, but it makes implementing the
    # iteration logic much simpler
    def __init__(
        self,
        left: Optional['CellLayout'],
        right: Optional['CellLayout'],
        split_orientation: Optional[Literal['h', 'v']],
        split_position: Optional[float],
    ):
        self.left = left
        self.right = right
        self.split_orientation = split_orientation
        self.split_position = split_position
        self.parent = None
        self.cell_loc = None
        self.cell_dim = None
        self.floating = None

    def _leftmost(self) -> 'CellLayout':
        if self.left is None:
            return self
        return self.left._leftmost()

    def _next(self) -> Optional['CellLayout']:
        if self.right is not None:
            return self.right._leftmost()
        elif self.parent is None:
            return None
        elif self.parent.left is self:
            return self.parent
        else:
            cur_node = self
            while (
                cur_node.parent is not None
                and cur_node.parent.right is cur_node
            ):
                cur_node = cur_node.parent
            return cur_node.parent

    def _next_leaf(self) -> Optional['CellLayout']:
        candidate = self._next()
        while candidate is not None and not isinstance(candidate, Cell):
            candidate = candidate._next()
        return candidate

    def __getitem__(self, index: int) -> 'CellLayout':
        for i, cell in enumerate(self):
            if i == index:
                return cell

    def __next__(self) -> 'CellLayout':
        return self._next_leaf()

    def __iter__(self):
        cur_cell = self._leftmost()
        while cur_cell is not None:
            yield cur_cell
            cur_cell = cur_cell._next_leaf()
        if self.floating is not None:
            for f in self.floating:
                yield from f

    def __repr__(self) -> str:
        return self._repr()

    def __len__(self) -> int:
        if self.parent is not None:
            raise ValueError(
                'Cannot get length of a non-root layout, use '
                'len(layout.root) instead'
            )
        return sum(1 for _ in self)

    def __mod__(self, other: 'CellLayout') -> 'CellLayoutSubstitution':
        return CellLayoutSubstitution(
            layout=self,
            substitute=other,
        )

    def __truediv__(self, other: 'CellLayout') -> 'CellLayoutHorizontalChain':
        if isinstance(other, CellLayoutHorizontalChain):
            return CellLayoutHorizontalChain(
                chain=[self] + other.chain
            )
        return CellLayoutHorizontalChain(
            chain=[self, other]
        )

    def __or__(
        self, other: Union['CellLayout', 'CellLayoutArgument']
    ) -> Union['CellLayoutVerticalChain', 'CellLayout']:
        if isinstance(other, CellLayoutArgument):
            return vsplit(other.argument, self, other.layout)
        elif isinstance(other, CellLayoutVerticalChain):
            return CellLayoutVerticalChain(
                chain=[self] + other.chain
            )
        return CellLayoutVerticalChain(
            chain=[self, other]
        )

    def __add__(self, other: 'CellLayout') -> 'CellLayoutFloat':
        """Note that this 'sum' is not commutative"""
        return CellLayoutFloat(
            anchor=self,
            floating=other,
        )

    def __mul__(self, other: 'CellLayout') -> 'CellLayout':
        """Note that this 'product' is not commutative"""
        return product(self, other)

    def __matmul__(self, other: int) -> Tuple['CellLayout', 'CellLayout']:
        return break_at(self, other)

    def __lshift__(self, other: Any):
        return CellLayoutArgument(
            layout=self,
            argument=other,
        )

    @property
    def root(self) -> 'CellLayout':
        """Get the root of the layout tree"""
        if self.parent is None:
            return self
        return self.parent.root

    @property
    def direct_size(self):
        """The direct size of a layout excludes any floating cells"""
        return sum(1 for e in self if e.root is self)

    @property
    def breakpoints_left(self):
        """Get valid breakpoints on the left side of the layout tree."""
        self.split_orientation
        if self.left.split_orientation == self.split_orientation:
            yield from self.left.breakpoints

    @property
    def breakpoints_right(self):
        """Get valid breakpoints on the right side of the layout tree."""
        if self.right.split_orientation == self.split_orientation:
            yield from self.right.breakpoints

    @property
    def breakpoints(self):
        yield from self.breakpoints_left
        yield self
        yield from self.breakpoints_right

    @property
    def is_left(self):
        return self.parent is not None and self.parent.left is self

    @property
    def is_right(self):
        return self.parent is not None and self.parent.right is self

    @property
    def count(self):
        left_count = self.left.count if self.left is not None else 0
        right_count = self.right.count if self.right is not None else 0
        return left_count + right_count

    def copy(self, **extra_params) -> 'CellLayout':
        left = self.left.copy() if self.left is not None else None
        right = self.right.copy() if self.right is not None else None
        floating = (
            [f.copy() for f in self.floating]
            if self.floating is not None
            else None
        )
        copy = self.__class__(
            left=left,
            right=right,
            split_orientation=self.split_orientation,
            split_position=self.split_position,
            **extra_params,
        )
        copy.cell_loc = self.cell_loc
        copy.cell_dim = self.cell_dim
        if left is not None:
            left.parent = copy
        if right is not None:
            right.parent = copy
        if floating is not None:
            copy.floating = floating
        return copy

    def partition(
        self,
        width: int,
        height: int,
        x_offset: int = 0,
        y_offset: int = 0,
        padding: int = 0,
    ) -> 'CellLayout':
        """
        Partition a canvas into a grid of cells specified by the layout
        """
        if self.split_orientation is None:
            assert self.split_position is None
        elif self.split_orientation == 'v':
            left_width = round(self.split_position * width)
            right_width = width - left_width
            left_height = right_height = height
            left_x_offset = x_offset
            right_x_offset = x_offset + left_width
            left_y_offset = right_y_offset = y_offset
        elif self.split_orientation == 'h':
            left_height = round(self.split_position * height)
            right_height = height - left_height
            left_width = right_width = width
            left_y_offset = y_offset
            right_y_offset = y_offset + left_height
            left_x_offset = right_x_offset = x_offset
        else:
            raise ValueError(
                f'Invalid split orientation: {self.split_orientation}'
            )
        if self.left is not None:
            self.left.partition(
                width=left_width,
                height=left_height,
                x_offset=left_x_offset,
                y_offset=left_y_offset,
                padding=padding,
            )
        if self.right is not None:
            self.right.partition(
                width=right_width,
                height=right_height,
                x_offset=right_x_offset,
                y_offset=right_y_offset,
                padding=padding,
            )
        if self.floating is not None:
            for f in self.floating:
                floating_width = round(f.float_dim_rel[0] * width)
                floating_height = round(f.float_dim_rel[1] * height)
                floating_x_offset = round(f.float_loc_rel[0] * width)
                floating_y_offset = round(f.float_loc_rel[1] * height)
                f.partition(
                    width=floating_width,
                    height=floating_height,
                    x_offset=floating_x_offset,
                    y_offset=floating_y_offset,
                    padding=padding,
                )
        self.cell_loc = (x_offset + padding, y_offset + padding)
        self.cell_dim = (width - 2 * padding, height - 2 * padding)
        return self

    # def clear_partition(self):
    #     """
    #     Clear partition information from the layout
    #     """
    #     self.cell_loc = None
    #     self.cell_dim = None
    #     if self.left is not None:
    #         self.left.clear_partition()
    #     if self.right is not None:
    #         self.right.clear_partition()
    #     if self.floating is not None:
    #         for f in self.floating:
    #             f.clear_partition()

    # @property
    # def loc(self):
    #     if self.cell_loc is not None:
    #         return self.cell_loc
    #     else:
    #         left = self.left.loc if self.left is not None else None
    #         right = self.right.loc if self.right is not None else None
    #         if right is None:
    #             return left
    #         elif left is None:
    #             return right
    #         else:
    #             return (
    #                 min(left[0], right[0]),
    #                 min(left[1], right[1]),
    #             )

    # @property
    # def dim(self):
    #     if self.cell_dim is not None:
    #         return self.cell_dim
    #     else:
    #         left = self.left.dim if self.left is not None else None
    #         right = self.right.dim if self.right is not None else None
    #         if right is None:
    #             return left
    #         elif left is None:
    #             return right
    #         else:
    #             return (
    #                 max(left[0], right[0]),
    #                 max(left[1], right[1]),
    #             )

    def annotate(
        self,
        annotations: Mapping[int, Optional[Mapping]],
    ) -> 'AnnotatedLayout':
        """
        Annotate a layout with a dictionary of annotations
        """
        return AnnotatedLayout(
            layout=self.copy(),
            annotations=annotations,
        )

    def _repr(self, inject: Optional[str] = None) -> str:
        inject = '' if inject is None else inject
        if self.floating:
            floating = f', floating={self.floating}'
        else:
            floating = ''
        return (
            f'{self.__class__.__name__}'
            f'(left={self.left}, right={self.right}{floating}, '
            f'orientation={self.split_orientation}, '
            f'position={self.split_position}'
            f'{inject})'
        )


class Cell(CellLayout):
    """
    Leaf cell in a binary layout tree
    """
    def __init__(self):
        super().__init__(
            left=None,
            right=None,
            split_orientation=None,
            split_position=None
        )

    def __repr__(self) -> str:
        return f'Cell(loc={self.cell_loc}, dim={self.cell_dim})'

    def copy(self) -> 'Cell':
        return self.__class__()

    @property
    def index(self):
        return list(self.root).index(self)

    @property
    def root_loc(self):
        # Neither the most efficient nor the most flexible way to do this,
        # but we'll take it for now
        loc = self.root.copy().partition(BIG, BIG)[self.index].cell_loc
        return (loc[0] / BIG, loc[1] / BIG)

    @property
    def root_dim(self):
        # Neither the most efficient nor the most flexible way to do this,
        # but we'll take it for now
        dim = self.root.copy().partition(BIG, BIG)[self.index].cell_dim
        return (dim[0] / BIG, dim[1] / BIG)

    @property
    def count(self):
        return 1


class FloatingCellLayout(CellLayout):
    """
    A cell layout that floats on top of other cells
    """
    def __init__(
        self,
        float_loc_rel: Tuple[float, float],
        float_dim_rel: Tuple[float, float],
        left: Optional['CellLayout'],
        right: Optional['CellLayout'],
        split_orientation: Optional[Literal['h', 'v']],
        split_position: Optional[float],
    ):
        self.left = left
        self.right = right
        self.split_orientation = split_orientation
        self.split_position = split_position
        self.parent = None
        self.cell_loc = None
        self.cell_dim = None
        self.float_loc_rel = float_loc_rel
        self.float_dim_rel = float_dim_rel
        self.floating = None

    def __repr__(self) -> str:
        return self._repr(
            f', float_loc_rel={self.float_loc_rel}, '
            f'float_dim_rel={self.float_dim_rel}'
        )

    def copy(self, **extra_params) -> 'FloatingCellLayout':
        return super().copy(
            float_loc_rel=self.float_loc_rel,
            float_dim_rel=self.float_dim_rel,
            **extra_params,
        )

    def defloat(self) -> 'CellLayout':
        if len(self) == 1:
            return Cell()
        else:
            new = self.copy()
            return CellLayout(
                left=new.left,
                right=new.right,
                split_orientation=self.split_orientation,
                split_position=self.split_position,
            )


class FloatingCell(Cell):
    def __init__(
        self,
        float_loc_rel: Tuple[float, float],
        float_dim_rel: Tuple[float, float],
    ):
        super().__init__()
        self.float_loc_rel = float_loc_rel
        self.float_dim_rel = float_dim_rel

    def __repr__(self) -> str:
        return (
            'FloatingCell('
            f'loc={self.cell_loc}, dim={self.cell_dim}, '
            f'float_loc_rel={self.float_loc_rel}, '
            f'float_dim_rel={self.float_dim_rel}'
            ')'
        )

    def copy(self, **extra_params) -> 'FloatingCell':
        return self.__class__(
            float_loc_rel=self.float_loc_rel,
            float_dim_rel=self.float_dim_rel,
        )

    def defloat(self) -> 'Cell':
        return Cell()


@dataclasses.dataclass
class AnnotatedLayout(CellLayout):
    """
    Layout with annotations
    """
    layout: CellLayout
    annotations: Mapping[int, Optional[Mapping]] = dataclasses.field(
        default_factory=dict
    )
    assigned: Optional[Sequence[bool]] = None

    def __post_init__(self):
        if self.assigned is None:
            self.assigned = [False] * len(self.layout)
        assert len(self.layout) == len(self.assigned)

    def _leftmost(self) -> CellLayout:
        return self.layout._leftmost()

    def _next(self) -> Optional[CellLayout]:
        return self.layout._next()

    def _next_leaf(self) -> Optional[CellLayout]:
        return self.layout._next_leaf()

    def __getitem__(self, index: int) -> CellLayout:
        return self.layout[index]

    def __next__(self) -> CellLayout:
        return self.layout.__next__()

    def __iter__(self):
        return self.layout.__iter__()

    def __repr__(self) -> str:
        return (
            f'AnnotatedLayout(layout={self.layout}, '
            f'annotations={self.annotations})'
        )

    def __len__(self) -> int:
        return len(self.layout)

    @property
    def root(self) -> 'CellLayout':
        raise NotImplementedError(
            'AnnotatedLayout does not support root property'
        )

    @property
    def direct_size(self):
        """The direct size of a layout excludes any floating cells"""
        return sum(1 for e in self.layout if e.root is self.layout)

    @property
    def count(self):
        return self.layout.count

    @property
    def floating(self):
        return self.layout.floating

    @cached_property
    def decomposition(self):
        if self.floating is None:
            return self, []
        else:
            max_annotation = self.direct_size
            root_layout = self.layout.copy()
            root_layout.floating = None
            root_layout = AnnotatedLayout(
                layout=root_layout,
                annotations={
                    i: annotation
                    for i, annotation in self.annotations.items()
                    if i < max_annotation
                },
            )
            floating_layouts = []
            for floating in self.floating:
                min_annotation = max_annotation
                max_annotation += len(floating)
                floating_layouts.append(
                    AnnotatedLayout(
                        layout=floating.defloat(),
                        annotations={
                            i - min_annotation: annotation
                            for i, annotation in self.annotations.items()
                            if i < max_annotation and i >= min_annotation
                        },
                    )
                )
            return root_layout, floating_layouts

    def partition(
        self,
        width: int,
        height: int,
        x_offset: int = 0,
        y_offset: int = 0,
        padding: int = 0,
    ) -> 'AnnotatedLayout':
        layout = self.layout.partition(
            width=width,
            height=height,
            x_offset=x_offset,
            y_offset=y_offset,
            padding=padding,
        )
        return AnnotatedLayout(
            layout=layout,
            annotations=self.annotations,
            assigned=self.assigned,
        )

    def annotate_cell(
        self,
        index: int,
        annotation: Optional[Mapping] = None,
        collision_policy: Literal[
            'error', 'overwrite', 'coalesce', 'ignore'
        ] = 'coalesce',
    ) -> 'AnnotatedLayout':
        if annotation is None:
            annotation = {}
        assert index < len(self), f'Invalid cell index: {index}'
        current_annotation = self.annotations.get(index, None)
        if current_annotation is not None:
            if collision_policy == 'error':
                raise ValueError(
                    f'Cell {index} already has an annotation: '
                    f'{current_annotation}'
                )
            elif collision_policy == 'overwrite':
                pass
            elif collision_policy == 'coalesce':
                annotation = {**current_annotation, **annotation}
            elif collision_policy == 'ignore':
                return self
        annotations = {**self.annotations, **{index: annotation}}
        return AnnotatedLayout(
            layout=self.layout,
            annotations=annotations,
            assigned=self.assigned,
        )

    def match_and_assign(
        self,
        query: Mapping,
        match_to_unannotated: bool = False,
    ) -> Tuple['AnnotatedLayout', int]:
        """
        Match annotations against a query and assign the first available cell
        to the query
        """
        assigned = self.assigned.copy()
        candidates = []
        for index, annotation in self.annotations.items():
            if assigned[index]:
                continue
            if (annotation is None) and match_to_unannotated:
                candidates += [(index, 0)]
            elif all(
                (
                    query.get(key, None) == value
                    or query.get(key, None) in value
                )
                for key, value in annotation.items()
            ):
                difficulty = len(annotation)
                candidates += [(index, difficulty)]
        # Select the candidate that is hardest to match
        if len(candidates) == 0:
            index = len(self)
        else:
            index, _ = sorted(
                candidates,
                key=lambda x: x[1],
                reverse=True,
            )[0]
            assigned[index] = True
        return AnnotatedLayout(
            layout=self.layout,
            annotations=self.annotations,
            assigned=assigned,
        ), index

    def match_and_assign_all(
        self,
        queries: Sequence[Mapping],
        force_unmatched: bool = False,
        # drop: Optional[Sequence[str]] = None,
    ) -> Tuple['AnnotatedLayout', Sequence[int]]:
        """
        Match annotations against a sequence of queries and assign the first
        available cell to each query
        """
        layout = self
        matched = [False] * len(queries)
        indices = [None] * len(queries)
        # if drop is not None:
        #     queries = [
        #         {k: v for k, v in query.items() if k not in drop}
        #         for query in queries
        #     ]
        for i, query in enumerate(queries):
            layout, index = layout.match_and_assign(
                query, match_to_unannotated=False
            )
            if index < len(layout):
                indices[i] = index
                matched[i] = True
        for i, query in enumerate(queries):
            if matched[i]:
                continue
            layout, index = layout.match_and_assign(
                query, match_to_unannotated=True
            )
            if index < len(layout):
                indices[i] = index
                matched[i] = True
            elif force_unmatched:
                assigned = layout.assigned.copy()
                try:
                    indices[i] = assigned.index(False)
                except ValueError:
                    break
                matched[i] = True
                assigned[indices[i]] = True
                layout = AnnotatedLayout(
                    layout=layout.layout,
                    annotations=layout.annotations,
                    assigned=assigned,
                )
        return layout, indices


@dataclasses.dataclass(frozen=True)
class GroupSpec:
    variable: str
    n_rows: int = -1
    n_cols: int = -1
    max_levels: Optional[int] = None
    order: Literal['row', 'col'] = 'row'

    def __post_init__(self):
        if self.n_rows == -1 and self.n_cols == -1:
            raise ValueError(
                'Must specify either n_row or n_col when creating a GroupSpec'
            )
        if self.max_levels is not None:
            assert self.max_levels % self.n_rows == 0, (
                f'Maximum number of levels ({self.max_levels}) must be '
                f'divisible by n_rows ({self.n_rows})'
            )
            assert self.max_levels % self.n_cols == 0, (
                f'Maximum number of levels ({self.max_levels}) must be '
                f'divisible by n_cols ({self.n_cols})'
            )
            if self.n_rows != -1 and self.n_cols != -1:
                assert self.max_levels == self.n_rows * self.n_cols, (
                    f'Maximum number of levels ({self.max_levels}) must be '
                    f'equal to n_rows ({self.n_rows}) * n_cols '
                    f'({self.n_cols}) when both are specified'
                )

    def __call__(self, metadata: Sequence[Mapping[str, str]]) -> Tuple[
        AnnotatedLayout,
        Optional[int]
    ]:
        seen = set()
        levels = []
        for obs in metadata:
            value = obs.get(self.variable, None)
            if value not in seen and value is not None:
                seen.add(value)
                levels.append(value)
        n_levels = len(levels)
        n_rows, n_cols = self.n_rows, self.n_cols
        if n_rows == -1:
            n_rows = ceil(n_levels / n_cols)
        elif n_cols == -1:
            n_cols = ceil(n_levels / n_rows)
        elif n_levels < n_rows * n_cols:
            warnings.warn(
                f'n_rows ({n_rows}) * n_cols ({n_cols}) is insufficient to '
                f'accommodate {n_levels} levels; your data will be '
                'truncated. To circumvent this issue, set either n_rows or '
                'n_cols to -1'
            )
        if self.max_levels is not None:
            if self.order == 'col':
                bp = self.max_levels // n_rows
                nb = ceil(n_cols / (bp + 1))
                layout = kernel = grid(
                    n_rows=n_rows,
                    n_cols=bp,
                    order=self.order,
                    kernel=Cell,
                )
                i = 0
                while i < nb:
                    layout = kernel | layout
                    i += 1
                layout = layout << (1 / (nb + 1))
            else:
                bp = self.max_levels // n_cols
                nb = ceil(n_rows / (bp + 1))
                layout = kernel = grid(
                    n_rows=bp,
                    n_cols=n_cols,
                    order=self.order,
                    kernel=Cell,
                )
                i = 0
                while i < nb:
                    layout = kernel / layout
                    i += 1
                layout = layout << (1 / (nb + 1))
        else:
            bp = None
            nb = 1
            layout = grid(
                n_rows=n_rows,
                n_cols=n_cols,
                order=self.order,
                kernel=Cell,
            )
        return (
            layout.annotate({
                i: {self.variable: level}
                for i, level in enumerate(levels)
            }),
            GroupSpec(
                variable=self.variable,
                n_rows=n_rows,
                n_cols=n_cols,
                max_levels=None,
                order=self.order,
            ),
            bp,
            nb,
        )


@dataclasses.dataclass(frozen=True, init=False)
class RowGroupSpec(GroupSpec):
    variable: str
    max_levels: Optional[int] = None

    def __init__(self, variable: str, max_levels: Optional[int] = None):
        object.__setattr__(self, 'variable', variable)
        object.__setattr__(self, 'n_rows', 1)
        object.__setattr__(self, 'n_cols', -1)
        object.__setattr__(self, 'max_levels', max_levels)
        object.__setattr__(self, 'order', 'col')



@dataclasses.dataclass(frozen=True, init=False)
class ColGroupSpec(GroupSpec):
    variable: str
    max_levels: Optional[int] = None

    def __init__(self, variable: str, max_levels: Optional[int] = None):
        object.__setattr__(self, 'variable', variable)
        object.__setattr__(self, 'n_rows', -1)
        object.__setattr__(self, 'n_cols', 1)
        object.__setattr__(self, 'max_levels', max_levels)
        object.__setattr__(self, 'order', 'row')


def reindex_annotations(*cells: AnnotatedLayout) -> Mapping[int, Mapping]:
    aggregate_size = 0
    floating_size = 0
    root_annotations = {}
    floating_annotations = {}
    for cell in cells:
        direct_size = cell.direct_size
        root_annotations = {
            **root_annotations,
            **{
                aggregate_size + i: annotation
                for i, annotation in cell.annotations.items()
                if i < direct_size
            },
        }
        floating_annotations = {
            **floating_annotations,
            **{
                floating_size + i - direct_size: annotation
                for i, annotation in cell.annotations.items()
                if i >= direct_size
            },
        }
        aggregate_size += direct_size
        floating_size += len(cell) - direct_size
    floating_annotations = {
        aggregate_size + i: annotation
        for i, annotation in floating_annotations.items()
    }
    return {
        **root_annotations,
        **floating_annotations,
    }


@singledispatch
def _split(
    *cells: CellLayout,
    split_orientation: Literal['h', 'v'],
    split_position: float,
) -> CellLayout:
    if len(cells) == 1:
        return cells[0]
    elif len(cells) == 2:
        left, right = cells
        left = left.copy() if left is not None else None
        right = right.copy() if right is not None else None
        left_floating = left.floating or []
        right_floating = right.floating or []
        left.floating = None
        right.floating = None
        layout = CellLayout(
            left=left,
            right=right,
            split_orientation=split_orientation,
            split_position=split_position,
        )
        if left is not None:
            left.parent = layout
            left_floating = [
                refloat_layout(
                    e,
                    layout,
                    loc_rel=(0, 0),
                    dim_rel=(
                        (1, split_position)
                        if split_orientation == 'h'
                        else (split_position, 1)
                    ),
                    inner_loc_rel=e.float_loc_rel,
                    inner_dim_rel=e.float_dim_rel,
                )
                for e in left_floating
            ]
        if right is not None:
            right.parent = layout
            right_floating = [
                refloat_layout(
                    e,
                    layout,
                    loc_rel=(
                        (0, split_position)
                        if split_orientation == 'h'
                        else (split_position, 0)
                    ),
                    dim_rel=(
                        (1, 1 - split_position)
                        if split_orientation == 'h'
                        else (1 - split_position, 1)
                    ),
                    inner_loc_rel=e.float_loc_rel,
                    inner_dim_rel=e.float_dim_rel,
                )
                for e in right_floating
            ]
        floating = left_floating + right_floating or None
        layout.floating = floating
        return layout
    else:
        new_position = split_position / (1 - split_position)
        return split(
            split_orientation,
            split_position,
            cells[0],
            split(split_orientation, new_position, *cells[1:])
        )


@_split.register
def _(
    *cells: AnnotatedLayout,
    split_orientation: Literal['h', 'v'],
    split_position: float,
) -> AnnotatedLayout:
    layouts = [cell.layout for cell in cells]
    layout = _split(
        *layouts,
        split_orientation=split_orientation,
        split_position=split_position,
    )
    annotations = reindex_annotations(*cells)
    return AnnotatedLayout(
        layout=layout,
        annotations=annotations,
    )


def split(
    orientation: Literal['h', 'v'],
    position: float,
    *cells: CellLayout,
) -> CellLayout:
    """
    Create a split layout
    """
    return _split(
        *cells,
        split_orientation=orientation,
        split_position=position,
    )


def vsplit(
    position: float,
    *cells: CellLayout,
) -> CellLayout:
    """
    Create a vertical split layout
    """
    return split('v', position, *cells)


def hsplit(
    position: float,
    *cells: CellLayout,
) -> CellLayout:
    """
    Create a horizontal split layout
    """
    return split('h', position, *cells)


def grid(
    n_rows: int,
    n_cols: int,
    order: Literal['row', 'col'] = 'row',
    kernel: callable = Cell,
) -> CellLayout:
    """
    Create a grid layout
    """
    if order == 'row':
        layout = hsplit(
            1 / n_rows,
            *[
                vsplit(
                    1 / n_cols,
                    *[
                        kernel()
                        for _ in range(n_cols)
                    ]
                )
                for _ in range(n_rows)
            ]
        )
    elif order == 'col':
        layout = vsplit(
            1 / n_cols,
            *[
                hsplit(
                    1 / n_rows,
                    *[
                        kernel()
                        for _ in range(n_rows)
                    ]
                )
                for _ in range(n_cols)
            ]
        )
    else:
        raise ValueError(f'Invalid grid order: {order}')
    return layout


def refloat_layout(
    floating: CellLayout,
    anchor: CellLayout,
    loc_rel: Tuple[float, float],
    dim_rel: Tuple[float, float],
    inner_loc_rel: Tuple[float, float],
    inner_dim_rel: Tuple[float, float],
) -> CellLayout:
    new_dim_rel = tuple(
        i * j for i, j in zip(inner_dim_rel, dim_rel)
    )
    new_loc_rel = tuple(
        i + j * k
        for i, j, k in zip(loc_rel, inner_loc_rel, dim_rel)
    )
    return float_layout(
        floating,
        anchor,
        loc_rel=new_loc_rel,
        dim_rel=new_dim_rel,
    ).floating[-1]


@singledispatch
def _float_layout(
    anchor: CellLayout,
    floating: CellLayout,
    loc_rel: Tuple[float, float],
    dim_rel: Tuple[float, float],
) -> CellLayout:
    refloated = []
    if floating.floating is not None:
        inner = floating.floating
        for e in inner:
            refloated.append(
                refloat_layout(
                    e,
                    anchor,
                    loc_rel=loc_rel,
                    dim_rel=dim_rel,
                    inner_loc_rel=e.float_loc_rel,
                    inner_dim_rel=e.float_dim_rel,
                )
            )
    to_float = floating.copy()
    to_float.floating = None
    if isinstance(to_float, Cell):
        floating = FloatingCell(
            float_loc_rel=loc_rel,
            float_dim_rel=dim_rel,
        )
    else:
        floating = FloatingCellLayout(
            float_loc_rel=loc_rel,
            float_dim_rel=dim_rel,
            left=to_float.left,
            right=to_float.right,
            split_orientation=to_float.split_orientation,
            split_position=to_float.split_position,
        )
    if floating.left is not None:
        floating.left.parent = floating
    if floating.right is not None:
        floating.right.parent = floating
    anchor = anchor.copy()
    if anchor.floating is None:
        anchor.floating = [floating] + refloated
    else:
        anchor.floating += [floating] + refloated
    return anchor


@_float_layout.register
def _(
    anchor: AnnotatedLayout,
    floating: AnnotatedLayout,
    loc_rel: Tuple[float, float],
    dim_rel: Tuple[float, float],
) -> AnnotatedLayout:
    layout = _float_layout(
        anchor.layout,
        floating.layout,
        loc_rel=loc_rel,
        dim_rel=dim_rel,
    )
    anchor_size = len(anchor)
    floating_annotations = {
        anchor_size + i: annotation
        for i, annotation in floating.annotations.items()
    }
    annotations = {
        **anchor.annotations,
        **floating_annotations,
    }
    return AnnotatedLayout(
        layout=layout,
        annotations=annotations,
    )


def float_layout(
    floating: CellLayout,
    anchor: CellLayout,
    loc_rel: Tuple[float, float],
    dim_rel: Tuple[float, float],
) -> CellLayout:
    return _float_layout(
        anchor,
        floating,
        loc_rel=loc_rel,
        dim_rel=dim_rel,
    )


@singledispatch
def _substitute(
    orig: CellLayout,
    substitute: CellLayout,
    index: int,
) -> CellLayout:
    if len(orig) <= 1:
        return substitute
    new = orig.copy()
    substitute = substitute.copy()

    cur_idx = 0
    cur_cell = new._leftmost()
    while cur_idx < index:
        cur_cell = cur_cell._next_leaf()
        cur_idx += 1
    cell_root_loc = cur_cell.root_loc
    cell_root_dim = cur_cell.root_dim
    if cur_cell.parent.left is cur_cell:
        cur_cell.parent.left = substitute
    else:
        cur_cell.parent.right = substitute
    substitute.parent = cur_cell.parent

    if substitute.floating is not None:
        substitute_floating = [
            refloat_layout(
                e,
                substitute,
                loc_rel=cell_root_loc,
                dim_rel=cell_root_dim,
                inner_loc_rel=e.float_loc_rel,
                inner_dim_rel=e.float_dim_rel,
            )
            for e in substitute.floating
        ]
        new.floating = (
            new.floating + substitute_floating
            if new.floating is not None
            else substitute_floating
        )
        substitute.floating = None
    return new


@_substitute.register
def _(
    orig: AnnotatedLayout,
    substitute: AnnotatedLayout,
    index: int,
) -> AnnotatedLayout:
    layout = _substitute(
        orig.layout,
        substitute.layout,
        index=index,
    )
    substitute_size = substitute.direct_size
    orig_size = orig.direct_size
    new_size = layout.direct_size
    incr = substitute_size - 1
    root_annotations = {
        i: annotation
        for i, annotation in orig.annotations.items()
        if (i < orig_size and i != index)
    }
    root_annotations = {
        **{
            i if i < index else i + incr: annotation
            for i, annotation in root_annotations.items()
        },
        **{
            i + index: annotation
            for i, annotation in substitute.annotations.items()
            if i < substitute_size
        },
    }
    if layout.floating is not None:
        orig_floating_size = len(orig) - orig_size
        floating_annotations = {
            **{
                i - orig_size + new_size: annotation
                for i, annotation in orig.annotations.items()
                if i >= orig_size
            },
            **{
                (
                    i - substitute_size + new_size + orig_floating_size
                ): annotation
                for i, annotation in substitute.annotations.items()
                if i >= substitute_size
            },
        }
    else:
        floating_annotations = {}
    annotations = {
        **root_annotations,
        **floating_annotations,
    }
    return AnnotatedLayout(
        layout=layout,
        annotations=annotations,
    )


def substitute(
    orig: CellLayout,
    substitute: CellLayout,
    index: int,
) -> CellLayout:
    """Substitute a cell in the layout with another layout"""
    return _substitute(
        orig,
        substitute,
        index=index,
    )


@singledispatch
def _product(
    outer: CellLayout,
    inner: CellLayout,
) -> CellLayout:
    if outer.floating is not None:
        # Try to lift the floating cells and recompose
        outer_base = outer.copy()
        outer_base.floating = None
        layout = product(outer_base, inner)
        float_loc_rel = [e.float_loc_rel for e in outer.floating]
        float_dim_rel = [e.float_dim_rel for e in outer.floating]
        floating_product = [
            product(e, inner)
            for e in outer.floating
        ]
        for floating, loc, dim in zip(
            floating_product,
            float_loc_rel,
            float_dim_rel,
        ):
            layout = float_layout(
                floating,
                layout,
                loc_rel=loc,
                dim_rel=dim,
            )
        return layout
    assert outer.floating is None, (
        'Floating cells are not supported in the outer layout'
    )
    layout = outer
    inner_size = inner.direct_size
    for i in range(len(layout)):
        layout = substitute(layout, inner, i * inner_size)
    return layout


@_product.register
def _(
    outer: AnnotatedLayout,
    inner: AnnotatedLayout
) -> AnnotatedLayout:
    if outer.floating is not None:
        # Try to lift the floating cells and recompose
        outer_base, outer_floating = outer.decomposition
        layout = product(outer_base, inner)
        float_loc_rel = [e.float_loc_rel for e in outer.floating]
        float_dim_rel = [e.float_dim_rel for e in outer.floating]
        floating_product = [
            product(e, inner)
            for e in outer_floating
        ]
        for floating, loc, dim in zip(
            floating_product,
            float_loc_rel,
            float_dim_rel,
        ):
            layout = float_layout(
                floating,
                layout,
                loc_rel=loc,
                dim_rel=dim,
            )
        return layout
    assert outer.floating is None, (
        'Floating cells are not supported in the outer layout'
    )
    layout = outer.layout * inner.layout
    outer_size = outer.direct_size
    inner_size = inner.direct_size
    fixed_size = outer_size * inner_size
    inner_total_size = len(inner)
    root_annotations = {
        i * inner_size + j: {
            **outer.annotations.get(i, {}),
            **inner.annotations.get(j, {})
        }
        for i in range(outer_size) for j in range(inner_size)
    }
    floating_annotations = {
        i * inner_total_size - (i + 1) * inner_size + j + fixed_size: {
            **outer.annotations.get(i, {}),
            **inner.annotations.get(j, {})
        }
        for i in range(outer_size)
        for j in range(inner_size, inner_total_size)
    }

    return AnnotatedLayout(
        layout=layout,
        annotations={
            **root_annotations,
            **floating_annotations,
        },
    )


def product(
    inner: CellLayout,
    outer: CellLayout,
) -> CellLayout:
    """Product of two layouts"""
    return _product(
        inner,
        outer,
    )


@singledispatch
def _break_at(
    layout: CellLayout,
    index: int,
) -> Tuple[CellLayout, CellLayout]:
    to_break = layout.copy()
    broken = None
    breakpoints_left = list(to_break.breakpoints_left)
    breakpoints_right = list(to_break.breakpoints_right)
    breakpoints = (
        breakpoints_left + [to_break] + breakpoints_right
    )
    break_point = breakpoints[index]
    breakpoints = dict(zip(breakpoints, range(len(breakpoints))))
    if break_point is to_break:
        left = to_break.left
        right = to_break.right
        left.parent = None
        right.parent = None
        return left, right

    # By the way we've defined valid break points, we can only traverse
    # valid break points when moving from the root to any valid break
    # point. This means that we can use the index of the break point
    # to determine whether we need to traverse left or right to get to
    # it.
    pointer = to_break
    leftward = index < breakpoints[pointer]
    parent = None
    break_leftward = None
    while pointer != break_point:
        previous = pointer
        pointer = pointer.left if leftward else pointer.right
        if ((index < breakpoints[pointer]) != leftward):
            if leftward:
                previous.left = None
            else:
                previous.right = None
            pointer.parent = None
            if broken is None:
                break_leftward = leftward
                broken = pointer
            elif parent is not None:
                break_leftward = leftward
                if leftward:
                    parent.right = pointer
                else:
                    parent.left = pointer
                pointer.parent = parent
            leftward = not leftward
            parent = previous
    if broken is None:
        parent = previous
        break_leftward = leftward
        broken = pointer

    attach_left = breakpoints[parent] < index
    branch_upper = (
        break_point.left if attach_left else break_point.right
    )
    branch_lower = (
        break_point.right if attach_left else break_point.left
    )
    # break_point.left = None
    # break_point.right = None
    branch_upper.parent = None
    branch_lower.parent = None
    match break_leftward:
        case True:
            parent.left = branch_upper
            branch_upper.parent = parent
            if break_point.parent is None or break_point.parent is parent:
                broken = branch_lower
            elif break_point.is_left:
                break_point.parent.left = branch_lower
                branch_lower.parent = break_point.parent
            elif break_point.is_right:
                break_point.parent.right = branch_lower
                branch_lower.parent = break_point.parent
        case False:
            parent.right = branch_upper
            branch_upper.parent = parent
            if break_point.parent is None or break_point.parent is parent:
                broken = branch_lower
            elif break_point.is_left:
                break_point.parent.left = branch_lower
                branch_lower.parent = break_point.parent
            elif break_point.is_right:
                break_point.parent.right = branch_lower
                branch_lower.parent = break_point.parent
    #break_point.parent = None
    if index < breakpoints[to_break]:
        return broken, to_break
    else:
        return to_break, broken


@_break_at.register
def _(
    layout: AnnotatedLayout,
    index: int,
) -> Tuple[AnnotatedLayout, AnnotatedLayout]:
    left, right = _break_at(
        layout.layout,
        index=index,
    )
    left_size = left.direct_size
    right_size = right.direct_size
    left_annotations = {
        i: annotation
        for i, annotation in layout.annotations.items()
        if i < left_size
    }
    right_annotations = {
        i - left_size: annotation
        for i, annotation in layout.annotations.items()
        if i >= left_size and i < left_size + right_size
    }
    return (
        AnnotatedLayout(
            layout=left,
            annotations=left_annotations,
        ),
        AnnotatedLayout(
            layout=right,
            annotations=right_annotations,
        ),
    )


def break_at(
    layout: CellLayout,
    index: int,
) -> Tuple[CellLayout, CellLayout]:
    """Break a layout at a given index"""
    return _break_at(
        layout,
        index=index,
    )



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Tests that require an interactive plotter display window
"""
import pytest
import templateflow.api as tflow
from hyve_examples import get_pain_thresh_nifti
from hyve.flows import plotdef
from hyve.transforms import (
    plot_to_display,
    plot_final_image,
    save_snapshots,
    surf_from_archive,
    surf_scalars_from_nifti,
    points_scalars_from_nifti,
)


@pytest.mark.ci_unsupported
def test_plotter_flow_syntax():
    plot_f = plotdef(
        surf_from_archive(),
        surf_scalars_from_nifti('gmdensity', template='fsaverage'),
        points_scalars_from_nifti('pain'),
        plot_to_display(),
    )
    plot_f(
        template='fsaverage',
        load_mask=True,
        gmdensity_nifti=tflow.get(
            template='MNI152NLin2009cAsym',
            suffix='probseg',
            label='GM',
            resolution=2
        ),
        pain_nifti=get_pain_thresh_nifti(),
        surf_projection=('pial',),
        surf_alpha=0.5,
    )


@pytest.mark.ci_unsupported
def test_plotter_final_capture():
    plot_f = plotdef(
        surf_from_archive(),
        surf_scalars_from_nifti('gmdensity', template='fsaverage'),
        points_scalars_from_nifti('pain'),
        plot_final_image(n_scenes=1), # n_scenes > 1 is not supported yet
        save_snapshots(
            fname_spec=(
                'scalars-{surfscalars}_hemisphere-{hemisphere}_view-{view}'
            ),
        ),
    )
    plot_f(
        template='fsaverage',
        load_mask=True,
        gmdensity_nifti=tflow.get(
            template='MNI152NLin2009cAsym',
            suffix='probseg',
            label='GM',
            resolution=2
        ),
        pain_nifti=get_pain_thresh_nifti(),
        surf_projection=('pial',),
        surf_alpha=0.5,
        output_dir='/tmp',
    )



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Unit tests for elementary brain network visualisations
"""
import pytest
import numpy as np
import pandas as pd
import pyvista as pv

from hyve_examples import (
    get_schaefer400_synthetic_conmat,
    get_schaefer400_cifti,
)
from hyve.flows import plotdef
from hyve.flows import add_network_data
from hyve.transforms import (
    surf_from_archive,
    surf_scalars_from_cifti,
    parcellate_colormap,
    add_node_variable,
    add_edge_variable,
    plot_to_image,
    save_snapshots,
    save_grid,
    node_coor_from_parcels,
    build_network,
    add_network_overlay,
)


def test_net():
    parcellation = get_schaefer400_cifti()
    cov = pd.read_csv(
        get_schaefer400_synthetic_conmat(), sep='\t', header=None
    ).values

    vis_nodes_edge_selection = np.zeros(400, dtype=bool)
    vis_nodes_edge_selection[0:5] = True
    vis_nodes_edge_selection[200:205] = True

    plot_f = plotdef(
        surf_from_archive(),
        surf_scalars_from_cifti('parcellation', plot=False),
        add_network_data(
            add_node_variable('vis'),
            add_edge_variable(
                "vis_conn",
                threshold=10,
                topk_threshold_nodewise=True,
                absolute=True,
                incident_node_selection=vis_nodes_edge_selection,
                emit_degree=True,
            ),
            add_edge_variable(
                "vis_internal_conn",
                absolute=True,
                connected_node_selection=vis_nodes_edge_selection,
            ),
        ),
        node_coor_from_parcels('parcellation'),
        build_network('vis'),
        parcellate_colormap('parcellation', 'network', target='node'),
        plot_to_image(),
        save_snapshots(
            fname_spec=(
                'network-schaefer400_view-{view}'
            ),
        ),
    )
    plot_f(
        template='fsLR',
        surf_projection='inflated',
        surf_alpha=0.2,
        parcellation_cifti=parcellation,
        node_radius='vis_conn_degree',
        node_color='index',
        edge_color='vis_conn_sgn',
        edge_radius='vis_conn_val',
        vis_nodal=vis_nodes_edge_selection.astype(int),
        vis_conn_adjacency=cov,
        vis_internal_conn_adjacency=cov,
        views=('dorsal', 'left', 'posterior'),
        output_dir='/tmp',
    )


@pytest.mark.parametrize('cmap', ['network', 'modal'])
def test_net_highlight(cmap):
    parcellation = get_schaefer400_cifti()
    cov = pd.read_csv(
        get_schaefer400_synthetic_conmat(), sep='\t', header=None
    ).values

    vis_nodes_edge_selection = np.zeros(400, dtype=bool)
    vis_nodes_edge_selection[0:2] = True
    vis_nodes_edge_selection[200:202] = True

    plot_f = plotdef(
        surf_from_archive(),
        surf_scalars_from_cifti('parcellation', plot=False),
        #build_network('vis'),
        add_network_overlay(
            'vis',
            add_network_data(
                add_edge_variable(
                    "vis_conn",
                    absolute=True,
                    incident_node_selection=vis_nodes_edge_selection,
                ),
                add_edge_variable(
                    "vis_internal_conn",
                    absolute=True,
                    connected_node_selection=vis_nodes_edge_selection,
                    emit_degree=True,
                    emit_incident_nodes=(0.2, 1),
                    removed_val=0.03,
                    surviving_val=1.0,
                ),
                how='left',
            ),
            parcellate_colormap('parcellation', 'modal', target='node'),
            node_coor_from_parcels('parcellation'),
        ),
        plot_to_image(),
        save_snapshots(
            fname_spec=(
                f'network-schaefer400_desc-visual_cmap-{cmap}_view-{{view}}'
            ),
        ),
    )
    plot_f(
        template="fsLR",
        surf_projection='veryinflated',
        surf_alpha=0.2,
        parcellation_cifti=parcellation,
        surf_scalars_cmap=cmap,
        vis_conn_adjacency=cov,
        vis_internal_conn_adjacency=cov,
        vis_node_radius='vis_internal_conn_degree',
        vis_node_color='index',
        vis_node_alpha='vis_internal_conn_incidents',
        vis_edge_color='vis_conn_sgn',
        vis_edge_radius='vis_conn_val',
        vis_edge_alpha='vis_internal_conn_val',
        views=("dorsal", "left", "posterior"),
        output_dir='/tmp',
    )


def test_net_highlight_nooverlay():
    parcellation = get_schaefer400_cifti()
    cov = pd.read_csv(
        get_schaefer400_synthetic_conmat(), sep='\t', header=None
    ).values

    vis_nodes_edge_selection = np.zeros(400, dtype=bool)
    vis_nodes_edge_selection[0:2] = True
    vis_nodes_edge_selection[200:202] = True

    plot_f = plotdef(
        surf_from_archive(),
        surf_scalars_from_cifti('parcellation', plot=False),
        #build_network('vis'),
        add_network_data(
            add_edge_variable(
                "vis_conn",
                absolute=True,
                incident_node_selection=vis_nodes_edge_selection,
            ),
            add_edge_variable(
                "vis_internal_conn",
                absolute=True,
                connected_node_selection=vis_nodes_edge_selection,
                emit_degree=True,
                emit_incident_nodes=(0.2, 1),
                removed_val=0.03,
                surviving_val=1.0,
            ),
            how='left',
        ),
        node_coor_from_parcels('parcellation'),
        build_network('vis'),
        parcellate_colormap('parcellation', 'modal', target='node'),
        plot_to_image(),
        save_snapshots(
            fname_spec=(
                'network-schaefer400_desc-visual_view-{view}'
            ),
        ),
    )
    plot_f(
        template="fsLR",
        surf_projection='veryinflated',
        surf_alpha=0.2,
        parcellation_cifti=parcellation,
        vis_conn_adjacency=cov,
        vis_internal_conn_adjacency=cov,
        node_radius='vis_internal_conn_degree',
        node_color='index',
        node_alpha='vis_internal_conn_incidents',
        edge_color='vis_conn_sgn',
        edge_radius='vis_conn_val',
        edge_alpha='vis_internal_conn_val',
        views=("dorsal", "left", "posterior"),
        output_dir='/tmp',
    )


def test_net_fig():
    parcellation = get_schaefer400_cifti()
    cov = pd.read_csv(
        get_schaefer400_synthetic_conmat(), sep='\t', header=None
    ).values

    vis_nodes_edge_selection = np.zeros(400, dtype=bool)
    vis_nodes_edge_selection[0:5] = True
    vis_nodes_edge_selection[200:205] = True

    plot_f = plotdef(
        surf_from_archive(),
        surf_scalars_from_cifti('parcellation', plot=False),
        add_network_data(
            add_node_variable('vis'),
            add_edge_variable(
                "vis_conn",
                threshold=10,
                topk_threshold_nodewise=True,
                absolute=True,
                incident_node_selection=vis_nodes_edge_selection,
                emit_degree=True,
            ),
            add_edge_variable(
                "vis_internal_conn",
                absolute=True,
                connected_node_selection=vis_nodes_edge_selection,
            ),
        ),
        node_coor_from_parcels('parcellation'),
        build_network('vis'),
        parcellate_colormap('parcellation', 'network', target='node'),
        plot_to_image(),
        save_grid(
            n_cols=3, n_rows=1,
            canvas_size=(1800, 500),
            canvas_color=(1, 1, 1),
            scalar_bar_action='collect',
        ),
    )
    plot_f(
        template='fsLR',
        surf_projection='inflated',
        surf_alpha=0.2,
        parcellation_cifti=parcellation,
        node_radius='vis_conn_degree',
        node_color='index',
        edge_color='vis_conn_sgn',
        edge_radius='vis_conn_val',
        vis_nodal=vis_nodes_edge_selection.astype(int),
        vis_conn_adjacency=cov,
        vis_internal_conn_adjacency=cov,
        views=('dorsal', 'left', 'posterior'),
        window_size=(1200, 1000),
        output_dir='/tmp',
        fname_spec=f'network-schaefer400_view-all_page-{{page}}',
    )


def test_net_highlight_fig():
    parcellation = get_schaefer400_cifti()
    cov = pd.read_csv(
        get_schaefer400_synthetic_conmat(), sep='\t', header=None
    ).values

    vis_nodes_edge_selection = np.zeros(400, dtype=bool)
    vis_nodes_edge_selection[0:2] = True
    vis_nodes_edge_selection[200:202] = True

    plot_f = plotdef(
        surf_from_archive(),
        surf_scalars_from_cifti('parcellation', plot=False),
        #build_network('vis'),
        add_network_overlay(
            'vis',
            add_network_data(
                add_edge_variable(
                    "vis_conn",
                    absolute=True,
                    incident_node_selection=vis_nodes_edge_selection,
                ),
                add_edge_variable(
                    "vis_internal_conn",
                    absolute=True,
                    connected_node_selection=vis_nodes_edge_selection,
                    emit_degree=True,
                    emit_incident_nodes=(0.2, 1),
                    removed_val=0.03,
                    surviving_val=1.0,
                ),
                how='left',
            ),
            parcellate_colormap('parcellation', 'modal', target='node'),
            node_coor_from_parcels('parcellation'),
        ),
        plot_to_image(),
        save_grid(
            n_cols=3, n_rows=1,
            canvas_size=(1800, 500),
            canvas_color=(1, 1, 1),
            fname_spec=f'network-schaefer400_desc-visual_view-all_page-{{page}}',
            scalar_bar_action='collect',
        ),
    )
    plot_f(
        template="fsLR",
        surf_projection='veryinflated',
        surf_alpha=0.2,
        parcellation_cifti=parcellation,
        vis_conn_adjacency=cov,
        vis_internal_conn_adjacency=cov,
        vis_node_radius='vis_internal_conn_degree',
        vis_node_color='index',
        vis_node_alpha='vis_internal_conn_incidents',
        vis_edge_color='vis_conn_sgn',
        vis_edge_radius='vis_conn_val',
        vis_edge_alpha='vis_internal_conn_val',
        views=("dorsal", "left", "posterior"),
        window_size=(1200, 1000),
        output_dir='/tmp',
    )



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Unit tests for elementary brain surface visualisations
"""
import pytest

import numpy as np
import templateflow.api as tflow

from hyve_examples import (
    get_null400_cifti,
    get_null400_gifti,
    get_poldrack_freesurfer,
)
from hyve.flows import plotdef
from hyve.transforms import (
    surf_from_archive,
    surf_from_freesurfer,
    surf_scalars_from_cifti,
    surf_scalars_from_freesurfer,
    surf_scalars_from_gifti,
    surf_scalars_from_nifti,
    parcellate_colormap,
    parcellate_surf_scalars,
    scatter_into_parcels,
    vertex_to_face,
    plot_to_html,
    plot_to_image,
    save_snapshots,
)


def test_scalars():
    plot_f = plotdef(
        surf_from_archive(),
        surf_scalars_from_nifti('gmdensity', template='fsaverage', plot=True),
        plot_to_image(),
        save_snapshots(
            fname_spec=(
                'scalars-{surfscalars}_hemisphere-{hemisphere}_view-{view}'
            ),
        ),
    )
    plot_f(
        template='fsaverage',
        load_mask=True,
        gmdensity_nifti=tflow.get(
            template='MNI152NLin2009cAsym',
            suffix='probseg',
            label='GM',
            resolution=2
        ),
        surf_projection=('pial',),
        hemisphere=['left', 'right'],
        output_dir='/tmp',
    )


@pytest.mark.ci_unsupported
def test_parcellation():
    plot_f = plotdef(
        surf_from_archive(),
        surf_scalars_from_cifti('parcellation', plot=True),
        parcellate_colormap('parcellation', 'network'),
        vertex_to_face('parcellation'),
        plot_to_image(),
        save_snapshots(
            fname_spec=(
                'scalars-{surfscalars}_hemisphere-{hemisphere}_view-{view}_cmap-network_mode-face'
            ),
        ),
    )
    plot_f(
        template="fsLR",
        load_mask=True,
        parcellation_cifti=get_null400_cifti(),
        surf_projection=('veryinflated',),
        hemisphere=['left', 'right'],
        views=[
            'lateral', 'medial', 'ventral', 'dorsal', 'anterior', 'posterior',
            (-20, 0, 0), ((65, 65, 0), (0, 0, 0), (0, 0, 1))
        ],
        output_dir='/tmp',
    )

    plot_f = plotdef(
        surf_from_archive(),
        surf_scalars_from_cifti('parcellation', plot=True),
        parcellate_colormap('parcellation', 'network'),
        plot_to_image(),
        save_snapshots(
            fname_spec=(
                'scalars-{surfscalars}_hemisphere-{hemisphere}_view-{view}_cmap-network_mode-vertex'
            ),
        ),
    )
    plot_f(
        template="fsLR",
        load_mask=True,
        parcellation_cifti=get_null400_cifti(),
        surf_projection=('veryinflated',),
        surf_scalars_boundary_color='black',
        surf_scalars_boundary_width=5,
        hemisphere=['left', 'right'],
        views=[
            'lateral', 'medial', 'ventral', 'dorsal', 'anterior', 'posterior',
            (-20, 0, 0), ((65, 65, 0), (0, 0, 0), (0, 0, 1))
        ],
        output_dir='/tmp',
    )


@pytest.mark.ci_unsupported
def test_parcellation_modal_cmap():
    plot_f = plotdef(
        surf_from_archive(),
        surf_scalars_from_gifti('parcellation', plot=True),
        parcellate_colormap('parcellation', 'modal'),
        vertex_to_face('parcellation'),
        plot_to_image(),
        save_snapshots(
            fname_spec=(
                'scalars-{surfscalars}_hemisphere-{hemisphere}_view-{view}_cmap-modal_mode-face'
            ),
        ),
    )
    parcellation_gifti = get_null400_gifti()
    plot_f(
        template="fsLR",
        load_mask=True,
        parcellation_gifti_left=parcellation_gifti['left'],
        parcellation_gifti_right=parcellation_gifti['right'],
        surf_projection=('veryinflated',),
        hemisphere=['left', 'right'],
        views=[
            'lateral', 'medial', 'ventral', 'dorsal', 'anterior', 'posterior',
            (-20, 0, 0), ((65, 65, 0), (0, 0, 0), (0, 0, 1))
        ],
        output_dir='/tmp',
    )

    plot_f = plotdef(
        surf_from_archive(),
        surf_scalars_from_cifti('parcellation', plot=True),
        parcellate_colormap('parcellation', 'modal'),
        plot_to_image(),
        save_snapshots(
            fname_spec=(
                'scalars-{surfscalars}_hemisphere-{hemisphere}_view-{view}_cmap-modal_mode-vertex'
            ),
        ),
    )
    plot_f(
        template="fsLR",
        load_mask=True,
        parcellation_cifti=get_null400_cifti(),
        surf_projection=('veryinflated',),
        surf_scalars_boundary_color='black',
        surf_scalars_boundary_width=5,
        hemisphere=['left', 'right'],
        views=[
            'lateral', 'medial', 'ventral', 'dorsal', 'anterior', 'posterior',
            (-20, 0, 0), ((65, 65, 0), (0, 0, 0), (0, 0, 1))
        ],
        output_dir='/tmp',
    )

@pytest.mark.ci_unsupported
def test_parcellation_html():
    plot_f = plotdef(
        surf_from_archive(),
        surf_scalars_from_cifti('parcellation', plot=True),
        parcellate_colormap('parcellation', 'network'),
        vertex_to_face('parcellation'),
        plot_to_html(),
    )
    plot_f(
        template="fsLR",
        load_mask=True,
        parcellation_cifti=get_null400_cifti(),
        surf_projection=['veryinflated'],
        # surf_scalars_boundary_color='black',
        # surf_scalars_boundary_width=5,
        hemisphere=['left', 'right'],
        window_size=(800, 800),
        output_dir='/tmp',
        fname_spec=(
            'scalars-{surfscalars}_hemisphere-{hemisphere}_cmap-network'
        ),
    )

@pytest.mark.ci_unsupported
def test_parcellated_scalars():
    plot_f = plotdef(
        surf_from_archive(),
        surf_scalars_from_nifti('gmdensity', template='fsLR', plot=False),
        surf_scalars_from_cifti('parcellation', plot=False),
        parcellate_surf_scalars('gmdensity', 'parcellation'),
        vertex_to_face('gmdensityParcellated', interpolation='mode'),
        plot_to_image(),
        save_snapshots(),
    )
    plot_f(
        template="fsLR",
        load_mask=True,
        parcellation_cifti=get_null400_cifti(),
        gmdensity_nifti=tflow.get(
            template='MNI152NLin2009cAsym',
            suffix='probseg',
            label="GM",
            resolution=2
        ),
        surf_projection=['inflated'],
        surf_scalars_clim=(0.1, 0.9),
        surf_scalars_below_color=(0, 0, 0, 0),
        hemisphere=['left', 'right'],
        output_dir='/tmp',
        fname_spec=(
            'scalars-{surfscalars}_hemisphere-{hemisphere}_view-{view}_parcellation-null'
        ),
    )

    parcellated = np.random.rand(400)
    plot_f = plotdef(
        surf_from_archive(),
        surf_scalars_from_cifti('parcellation', plot=False),
        scatter_into_parcels('noise', 'parcellation'),
        vertex_to_face('noise', interpolation='mode'),
        plot_to_image(),
        save_snapshots(
            fname_spec=(
                'scalars-{surfscalars}_hemisphere-{hemisphere}_view-{view}_parcellation-null'
            ),
        ),
    )
    plot_f(
        template="fsLR",
        load_mask=True,
        parcellation_cifti=get_null400_cifti(),
        noise_parcellated=parcellated,
        surf_projection=['inflated'],
        surf_scalars_clim=(0, 1),
        surf_scalars_cmap='inferno',
        surf_scalars_below_color=(0, 0, 0, 0),
        hemisphere=['left', 'right'],
        output_dir='/tmp',
    )


def test_freesurfer():
    fs = get_poldrack_freesurfer()
    geom_left, morph_left = fs['left']
    geom_right, morph_right = fs['right']
    plot_f = plotdef(
        surf_from_freesurfer(projection='inflated'),
        surf_scalars_from_freesurfer('curv'),
        plot_to_image(),
        save_snapshots(
            fname_spec=(
                'scalars-{surfscalars}_hemisphere-{hemisphere}_view-{view}'
            ),
        ),
    )
    plot_f(
        inflated_left_surf=geom_left,
        inflated_right_surf=geom_right,
        curv_morph_left=morph_left,
        curv_morph_right=morph_right,
        surf_scalars_cmap='RdYlBu_r',
        surf_scalars_clim=(-0.35, 0.35),
        hemisphere=['left', 'right'],
        output_dir='/tmp',
    )



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Unit tests for layout representation
"""
from hyve.layout import CellLayout, Cell, hsplit, vsplit, grid, float_layout


def test_simple_layout():
    layout = vsplit(1 / 3,
        hsplit(1 / 2,
            Cell(),
            vsplit(1 / 4,
                Cell(),
                Cell(),
            ),
        ),
        Cell(),
        Cell(),
    ).partition(120, 120)
    cells = list(layout)
    assert cells[0].cell_loc == (0, 0)
    assert cells[0].cell_dim == (40, 60)
    assert cells[1].cell_loc == (0, 60)
    assert cells[1].cell_dim == (10, 60)
    assert cells[2].cell_loc == (10, 60)
    assert cells[2].cell_dim == (30, 60)
    assert cells[3].cell_loc == (40, 0)
    assert cells[3].cell_dim == (40, 120)
    assert cells[4].cell_loc == (80, 0)
    assert cells[4].cell_dim == (40, 120)


def test_grid_layout():
    gridlayout = grid(n_cols=6, n_rows=5).partition(1200, 900)
    cells = list(gridlayout)
    assert all([cells[i].cell_dim == (200, 180) for i in range(30)])
    for i in range(5):
        for j in range(6):
            assert cells[6 * i + j].cell_loc == (200 * j, 180 * i)

    gridlayout = grid(n_cols=6, n_rows=5, order='col').partition(1200, 900)
    cells = list(gridlayout)
    assert all([cells[i].cell_dim == (200, 180) for i in range(30)])
    for i in range(5):
        for j in range(6):
            assert cells[5 * j + i].cell_loc == (200 * j, 180 * i)


def test_layout_substitute():
    # layout = (
    #     (Cell() | (1 / 2) | (Cell() /
    #                          (1 / 4) /
    #                          Cell())
    #     ) / (1 / 3) / (
    #         Cell(),
    #         Cell(),
    #     )
    # )
    # layout = (
    #     (Cell() | (Cell() /
    #                Cell() <<
    #                (1 / 4)) << (1 / 2)) /
    #      Cell() /
    #      Cell() <<
    #      (1 / 3)
    # )
    layout = ((
        Cell() /
        (Cell() | Cell() << (1 / 4)) <<
        (1 / 2))| Cell() | Cell() << (1 / 3)
    )
    layout_inner = Cell() / Cell() / Cell() << (1 / 4)
    layout_sub = (layout % layout_inner << 1).partition(120, 120)
    cells = list(layout_sub)
    assert cells[0].cell_loc == (0, 0)
    assert cells[0].cell_dim == (40, 60)
    assert cells[1].cell_loc == (0, 60)
    assert cells[1].cell_dim == (10, 15)
    assert cells[2].cell_loc == (0, 75)
    assert cells[2].cell_dim == (10, 15)
    assert cells[3].cell_loc == (0, 90)
    assert cells[3].cell_dim == (10, 30)
    assert cells[4].cell_loc == (10, 60)
    assert cells[4].cell_dim == (30, 60)
    assert cells[5].cell_loc == (40, 0)
    assert cells[5].cell_dim == (40, 120)
    assert cells[6].cell_loc == (80, 0)
    assert cells[6].cell_dim == (40, 120)

def test_layout_product():
    layout0 = Cell() / Cell() << (1 / 3)
    layout1 = Cell() | Cell() << (1 / 4)
    layout01 = (layout0 * layout1).partition(120, 120)
    cells = list(layout01)
    assert cells[0].cell_loc == (0, 0)
    assert cells[0].cell_dim == (30, 40)
    assert cells[1].cell_loc == (30, 0)
    assert cells[1].cell_dim == (90, 40)
    assert cells[2].cell_loc == (0, 40)
    assert cells[2].cell_dim == (30, 80)
    assert cells[3].cell_loc == (30, 40)
    assert cells[3].cell_dim == (90, 80)

    layout10 = (layout1 * layout0).partition(120, 120)
    cells = list(layout10)
    assert cells[0].cell_loc == (0, 0)
    assert cells[0].cell_dim == (30, 40)
    assert cells[1].cell_loc == (0, 40)
    assert cells[1].cell_dim == (30, 80)
    assert cells[2].cell_loc == (30, 0)
    assert cells[2].cell_dim == (90, 40)
    assert cells[3].cell_loc == (30, 40)
    assert cells[3].cell_dim == (90, 80)

    annotation0 = {
        0: {'x': 0},
        1: {'x': 1},
    }
    annotation1 = {
        0: {'y': 0},
        1: {'y': 1},
    }
    annotated0 = layout0.annotate(annotation0)
    annotated1 = layout1.annotate(annotation1)
    annotated01 = (annotated0 * annotated1).partition(120, 120)
    assert annotated01.annotations == {
        0: {'x': 0, 'y': 0},
        1: {'x': 0, 'y': 1},
        2: {'x': 1, 'y': 0},
        3: {'x': 1, 'y': 1},
    }
    for cell_a, cell_b in zip(annotated01.layout, layout01):
        assert cell_a.cell_loc == cell_b.cell_loc
        assert cell_a.cell_dim == cell_b.cell_dim

    annotated10 = (annotated1 * annotated0).partition(120, 120)
    assert annotated10.annotations == {
        0: {'x': 0, 'y': 0},
        1: {'x': 1, 'y': 0},
        2: {'x': 0, 'y': 1},
        3: {'x': 1, 'y': 1},
    }
    for cell_a, cell_b in zip(annotated10.layout, layout10):
        assert cell_a.cell_loc == cell_b.cell_loc
        assert cell_a.cell_dim == cell_b.cell_dim


def test_layout_floating():
    anchor = Cell() | Cell() | Cell() | Cell() << (1 / 5)
    floating = Cell() / Cell() << (1 / 3)
    floating_inner = Cell() | Cell() << (1 / 2)
    layout_inner = floating + floating_inner << ((0.1, 0.4), (0.8, 0.2))
    layout = anchor + layout_inner << ((0.7, 0.1), (0.2, 0.8))
    for i, cell in enumerate(layout):
        if i < 4:
            assert cell.root is layout
        elif i < 6:
            assert cell.root is layout.floating[0]
        else:
            assert cell.root is layout.floating[1]

    layout.partition(500, 500)
    cells = list(layout)
    assert cells[0].cell_loc == (0, 0)
    assert cells[0].cell_dim == (100, 500)
    assert cells[1].cell_loc == (100, 0)
    assert cells[1].cell_dim == (100, 500)
    assert cells[2].cell_loc == (200, 0)
    assert cells[2].cell_dim == (100, 500)
    assert cells[3].cell_loc == (300, 0)
    assert cells[3].cell_dim == (200, 500)

    # Total area is (100, 400)
    assert cells[4].cell_loc == (350, 50)
    assert cells[4].cell_dim == (100, 133)
    assert cells[5].cell_loc == (350, 183)
    assert cells[5].cell_dim == (100, 267)

    # Total area is (80, 80)
    assert cells[6].cell_loc == (360, 210)
    assert cells[6].cell_dim == (40, 80)
    assert cells[7].cell_loc == (400, 210)
    assert cells[7].cell_dim == (40, 80)

    anchor = Cell() | Cell() << (1 / 2)
    floating = Cell() / Cell() << (1 / 4)
    base = anchor + floating << ((0.1, 0.1), (0.8, 0.8))
    layout = (
        base | base << (1 / 2)) / (
        base | base << (1 / 2)) << (1 / 2)
    layout.partition(400, 400)
    cells = list(layout)
    def check_cells(cells):
        assert cells[0].cell_loc == (0, 0)
        assert cells[0].cell_dim == (100, 200)
        assert cells[1].cell_loc == (100, 0)
        assert cells[1].cell_dim == (100, 200)
        assert cells[2].cell_loc == (200, 0)
        assert cells[2].cell_dim == (100, 200)
        assert cells[3].cell_loc == (300, 0)
        assert cells[3].cell_dim == (100, 200)
        assert cells[4].cell_loc == (0, 200)
        assert cells[4].cell_dim == (100, 200)
        assert cells[5].cell_loc == (100, 200)
        assert cells[5].cell_dim == (100, 200)
        assert cells[6].cell_loc == (200, 200)
        assert cells[6].cell_dim == (100, 200)
        assert cells[7].cell_loc == (300, 200)
        assert cells[7].cell_dim == (100, 200)

        # Total area is (160, 160)
        assert cells[8].cell_loc == (20, 20)
        assert cells[8].cell_dim == (160, 40)
        assert cells[9].cell_loc == (20, 60)
        assert cells[9].cell_dim == (160, 120)
        assert cells[10].cell_loc == (220, 20)
        assert cells[10].cell_dim == (160, 40)
        assert cells[11].cell_loc == (220, 60)
        assert cells[11].cell_dim == (160, 120)
        assert cells[12].cell_loc == (20, 220)
        assert cells[12].cell_dim == (160, 40)
        assert cells[13].cell_loc == (20, 260)
        assert cells[13].cell_dim == (160, 120)
        assert cells[14].cell_loc == (220, 220)
        assert cells[14].cell_dim == (160, 40)
        assert cells[15].cell_loc == (220, 260)
        assert cells[15].cell_dim == (160, 120)

    check_cells(cells)

    annotated0 = base.annotate(
        {0: {'i': 0}, 1: {'i': 1}, 2: {'i': 8}, 3: {'i': 9}}
    )
    annotated1 = base.annotate(
        {0: {'i': 2}, 1: {'i': 3}, 2: {'i': 10}, 3: {'i': 11}}
    )
    annotated2 = base.annotate(
        {0: {'i': 4}, 1: {'i': 5}, 2: {'i': 12}, 3: {'i': 13}}
    )
    annotated3 = base.annotate(
        {0: {'i': 6}, 1: {'i': 7}, 2: {'i': 14}, 3: {'i': 15}}
    )
    annotated = (
        annotated0 | annotated1 << (1 / 2)) / (
        annotated2 | annotated3 << (1 / 2)) << (1 / 2)
    annotated.partition(400, 400)
    cells = list(annotated)
    check_cells(cells)
    assert len(annotated) == 16
    for i in range(len(annotated)):
        assert annotated.annotations[i]['i'] == i

    annotated0 = base.annotate(
        {0: {'i': 0}, 1: {'i': None}, 2: {'i': 3}, 3: {'i': 4}}
    )
    annotated1 = base.annotate(
        {0: {'i': 1}, 1: {'i': 2}, 2: {'i': 5}, 3: {'i': 6}}
    )
    annotated = annotated0 % annotated1 << 1
    annotated.partition(400, 400)
    assert len(annotated) == 7
    for i in range(len(annotated)):
        assert annotated.annotations[i]['i'] == i
        if i > 2:
            assert type(annotated[i].root).__name__ == 'FloatingCellLayout'
    cells = list(annotated)
    assert cells[0].cell_loc == (0, 0)
    assert cells[0].cell_dim == (200, 400)
    assert cells[1].cell_loc == (200, 0)
    assert cells[1].cell_dim == (100, 400)
    assert cells[2].cell_loc == (300, 0)
    assert cells[2].cell_dim == (100, 400)
    assert cells[3].cell_loc == (40, 40)
    assert cells[3].cell_dim == (320, 80)
    assert cells[4].cell_loc == (40, 120)
    assert cells[4].cell_dim == (320, 240)
    assert cells[5].cell_loc == (220, 40)
    assert cells[5].cell_dim == (160, 80)
    assert cells[6].cell_loc == (220, 120)
    assert cells[6].cell_dim == (160, 240)

    annotated_a0 = anchor.annotate({0: {'i': 0}, 1: {'i': 1}})
    annotated_a1 = anchor.annotate({0: {'i': 2}, 1: {'i': 3}})
    annotated_a2 = anchor.annotate({0: {'i': 4}, 1: {'i': 5}})
    annotated_a3 = anchor.annotate({0: {'i': 6}, 1: {'i': 7}})
    annotated_f0 = floating.annotate({0: {'i': 8}, 1: {'i': 9}})
    annotated_f1 = floating.annotate({0: {'i': 10}, 1: {'i': 11}})
    annotated_f2 = floating.annotate({0: {'i': 12}, 1: {'i': 13}})
    annotated_f3 = floating.annotate({0: {'i': 14}, 1: {'i': 15}})
    annotated0 = annotated_a0 + annotated_f0 << ((0.1, 0.1), (0.8, 0.8))
    annotated1 = annotated_a1 + annotated_f1 << ((0.1, 0.1), (0.8, 0.8))
    annotated2 = annotated_a2 + annotated_f2 << ((0.1, 0.1), (0.8, 0.8))
    annotated3 = annotated_a3 + annotated_f3 << ((0.1, 0.1), (0.8, 0.8))
    annotated = (
        (annotated0 | annotated1 << (1 / 2)) /
        (annotated2 | annotated3 << (1 / 2)) << (1 / 2)
    )
    annotated.partition(400, 400)
    cells = list(annotated)
    check_cells(cells)
    assert len(annotated) == 16
    for i in range(len(annotated)):
        assert annotated.annotations[i]['i'] == i

    layout0 = Cell() | Cell() << (1 / 2)
    layout1 = Cell() / Cell() << (1 / 4)
    onecell = Cell()
    annotated0 = layout0.annotate({0: {'x': 0}, 1: {'x': 1}})
    annotated1 = layout1.annotate({0: {'y': 0}, 1: {'y': 1}})
    annotated2 = onecell.annotate({0: {'float0': True}})
    annotated3 = onecell.annotate({0: {'float1': True}})
    annotated12 = annotated1 + annotated2 << ((0.1, 0.6), (0.8, 0.3))
    annotated03 = annotated0 + annotated3 << ((0.6, 0.1), (0.3, 0.8))
    annotated = annotated0 * annotated12
    annotated.partition(400, 400)
    cells = list(annotated)
    annotations = annotated.annotations
    assert cells[0].cell_loc == (0, 0)
    assert cells[0].cell_dim == (200, 100)
    assert annotations[0] == {'x': 0, 'y': 0}
    assert cells[1].cell_loc == (0, 100)
    assert cells[1].cell_dim == (200, 300)
    assert annotations[1] == {'x': 0, 'y': 1}
    assert cells[2].cell_loc == (200, 0)
    assert cells[2].cell_dim == (200, 100)
    assert annotations[2] == {'x': 1, 'y': 0}
    assert cells[3].cell_loc == (200, 100)
    assert cells[3].cell_dim == (200, 300)
    assert annotations[3] == {'x': 1, 'y': 1}
    assert cells[4].cell_loc == (20, 240)
    assert cells[4].cell_dim == (160, 120)
    assert annotations[4] == {'float0': True, 'x': 0}
    assert cells[5].cell_loc == (220, 240)
    assert cells[5].cell_dim == (160, 120)
    assert annotations[5] == {'float0': True, 'x': 1}

    # Really I don't think you should do this in practice, but it's possible
    annotated = annotated03 * annotated12
    annotations = annotated.annotations
    cells = list(annotated.partition(400, 400))
    assert cells[0].cell_loc == (0, 0)
    assert cells[0].cell_dim == (200, 100)
    assert annotations[0] == {'x': 0, 'y': 0}
    assert cells[1].cell_loc == (0, 100)
    assert cells[1].cell_dim == (200, 300)
    assert annotations[1] == {'x': 0, 'y': 1}
    assert cells[2].cell_loc == (200, 0)
    assert cells[2].cell_dim == (200, 100)
    assert annotations[2] == {'x': 1, 'y': 0}
    assert cells[3].cell_loc == (200, 100)
    assert cells[3].cell_dim == (200, 300)
    assert annotations[3] == {'x': 1, 'y': 1}
    assert cells[4].cell_loc == (20, 240)
    assert cells[4].cell_dim == (160, 120)
    assert annotations[4] == {'x': 0, 'float0': True}
    assert cells[5].cell_loc == (220, 240)
    assert cells[5].cell_dim == (160, 120)
    assert annotations[5] == {'x': 1, 'float0': True}
    # TODO: need to double check the below -- something here isn't right
    assert cells[6].cell_loc == (240, 40)
    assert cells[6].cell_dim == (120, 80)
    assert annotations[6] == {'y': 0, 'float1': True}
    assert cells[7].cell_loc == (240, 120)
    assert cells[7].cell_dim == (120, 240)
    assert annotations[7] == {'y': 1, 'float1': True}
    assert cells[8].cell_loc == (252, 232)
    assert cells[8].cell_dim == (96, 96)
    assert annotations[8] == {'float0': True, 'float1': True}


def test_break():
    layout0 = Cell() | Cell() | Cell() << (1 / 3)
    layout1 = Cell() | Cell() << (1 /2)
    layout2 = Cell() / Cell() << (1 / 2)
    layout = layout0 * layout1 * layout2
    left, right = layout @ 0
    assert len(left), len(right) == (2, 10)
    left, right = layout @ 1
    assert len(left), len(right) == (4, 8)
    left, right = layout @ 2
    assert len(left), len(right) == (6, 6)
    left, right = layout @ 3
    assert len(left), len(right) == (8, 4)
    left, right = layout @ 4
    assert len(left), len(right) == (10, 2)

    assert len(list(layout.breakpoints)) == 5
    assert len(list(layout[0].parent.breakpoints)) == 1
    assert len(list(layout[0].parent.parent.breakpoints)) == 1

    layout = layout1 * layout0 * layout2
    breakpoints = list(layout.breakpoints)
    bpdict = dict(zip(breakpoints, range(len(breakpoints))))
    # not intuitive, because of operator precedence
    assert breakpoints[0] is layout.left
    assert breakpoints[1] is layout.left.right
    assert breakpoints[2] is layout
    assert breakpoints[3] is layout.right
    assert breakpoints[4] is layout.right.right
    left, right = layout @ 0
    assert len(left), len(right) == (2, 10)
    left, right = layout @ 1
    assert len(left), len(right) == (4, 8)
    left, right = layout @ 2
    assert len(left), len(right) == (6, 6)
    left, right = layout @ 3
    assert len(left), len(right) == (8, 4)
    left, right = layout @ 4
    assert len(left), len(right) == (10, 2)

    layout = (
        layout1 * layout1 * layout1 * layout0 * layout1 * layout2 * layout1
    )
    l = len(layout)
    k = 4
    layout @ 35
    for i in range(47):
        left, right = layout @ i
        try:
            len(right)
        except Exception:
            print('very bad', i)
            continue
        assert (len(left), len(right)) == ((k * (i + 1), l - k * (i + 1)))

    annotated = layout.annotate({i: {'i': i} for i in range(len(layout))})
    for i in range(47):
        left, right = annotated @ i
        assert len(left), len(right) == (k * (i + 1), l - k * (i + 1))
        assert left.annotations == {j: {'i': j} for j in range(k * (i + 1))}
        assert right.annotations == {
            i: {'i': j} for i, j in enumerate(range(k * (i + 1), l))
        }



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Unit tests for flat map visualisations loaded from a GIfTI file
"""
import pytest

import templateflow.api as tflow

from hyve_examples import (
    get_null400_cifti,
    get_fsLR_flatmap_gifti,
)
from hyve.flows import plotdef
from hyve.transforms import (
    surf_from_gifti,
    surf_scalars_from_cifti,
    surf_scalars_from_nifti,
    parcellate_colormap,
    vertex_to_face,
    plot_to_image,
    save_snapshots,
)


lh_mask = tflow.get(
    template='fsLR',
    hemi='L',
    desc='nomedialwall',
    density='32k',
)
rh_mask = tflow.get(
    template='fsLR',
    hemi='R',
    desc='nomedialwall',
    density='32k',
)
surf = get_fsLR_flatmap_gifti()


@pytest.mark.ci_unsupported
def test_scalars():
    plot_f = plotdef(
        surf_from_gifti(projection='flat'),
        surf_scalars_from_nifti('GM Density', template='fsLR'),
        plot_to_image(),
        save_snapshots(
            fname_spec=(
                'scalars-{surfscalars}_hemisphere-{hemisphere}_projection-flat'
            ),
        ),
    )
    plot_f(
        flat_left_surf=surf['left'],
        flat_right_surf=surf['right'],
        flat_left_mask=lh_mask,
        flat_right_mask=rh_mask,
        gm_density_nifti=tflow.get(
            template='MNI152NLin2009cAsym',
            suffix='probseg',
            label='GM',
            resolution=2,
        ),
        hemisphere=['left', 'right'],
        views=['down'],
        output_dir='/tmp',
    )


@pytest.mark.parametrize(
    'cmap, cmap_name', [
        ('network', 'network'),
        ('modal', 'modal'),
        ('network', 'bone')
    ])
def test_parcellation(cmap, cmap_name):
    plot_f = plotdef(
        surf_from_gifti(projection='flat'),
        surf_scalars_from_cifti('parcellation', plot=True),
        parcellate_colormap('parcellation', cmap),
        vertex_to_face('parcellation'),
        plot_to_image(),
        save_snapshots(
            fname_spec=(
                'scalars-{surfscalars}_hemisphere-{hemisphere}_'
                f'cmap-{cmap_name}_projection-flat'
            ),
        ),
    )
    plot_f(
        flat_left_surf=surf['left'],
        flat_right_surf=surf['right'],
        flat_left_mask=lh_mask,
        flat_right_mask=rh_mask,
        parcellation_cifti=get_null400_cifti(),
        surf_scalars_cmap=cmap_name,
        hemisphere=['left', 'right'],
        views=['down'],
        output_dir='/tmp',
    )



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
2-dimensional plot elements
~~~~~~~~~~~~~~~~~~~~~~~~~~~
Operations for creating and tiling 2-dimensional plots and plot elements.
"""
import base64
import dataclasses
import io
import re
from abc import abstractmethod
from collections.abc import Mapping as MappingABC
from statistics import median
from typing import Any, Literal, Mapping, Optional, Sequence, Tuple

import matplotlib.pyplot as plt
import numpy as np
import svg
from matplotlib import cm, colors, patheffects
from matplotlib.figure import Figure
from PIL import Image

from .const import (
    RASTER_DEFAULT_BOUNDING_BOX_HEIGHT,
    RASTER_DEFAULT_BOUNDING_BOX_WIDTH,
    RASTER_DEFAULT_FORMAT,
    SCALAR_BAR_DEFAULT_BELOW_COLOR,
    SCALAR_BAR_DEFAULT_FONT,
    SCALAR_BAR_DEFAULT_FONT_COLOR,
    SCALAR_BAR_DEFAULT_FONT_OUTLINE_COLOR,
    SCALAR_BAR_DEFAULT_FONT_OUTLINE_MULTIPLIER,
    SCALAR_BAR_DEFAULT_LENGTH,
    SCALAR_BAR_DEFAULT_LIM_FONTSIZE_MULTIPLIER,
    SCALAR_BAR_DEFAULT_NAME,
    SCALAR_BAR_DEFAULT_NAME_FONTSIZE_MULTIPLIER,
    SCALAR_BAR_DEFAULT_NAME_SUFFIX,
    SCALAR_BAR_DEFAULT_NUM_SIG_FIGS,
    SCALAR_BAR_DEFAULT_ORIENTATION,
    SCALAR_BAR_DEFAULT_SPACING,
    SCALAR_BAR_DEFAULT_WIDTH,
    TEXT_DEFAULT_ANGLE,
    TEXT_DEFAULT_BOUNDING_BOX_HEIGHT,
    TEXT_DEFAULT_BOUNDING_BOX_WIDTH,
    TEXT_DEFAULT_CONTENT,
    TEXT_DEFAULT_FONT,
    TEXT_DEFAULT_FONT_COLOR,
    TEXT_DEFAULT_FONT_OUTLINE_COLOR,
    TEXT_DEFAULT_FONT_OUTLINE_MULTIPLIER,
    TEXT_DEFAULT_FONT_SIZE_MULTIPLIER,
    TYPICAL_DPI,
    Tensor,
)


@dataclasses.dataclass(frozen=True)
class ElementBuilder(MappingABC):
    """Addressable container for 2D plot elements."""

    def __getitem__(self, key: str):
        return self.__getattribute__(key)

    def __iter__(self):
        return iter(dataclasses.asdict(self))

    def __len__(self) -> int:
        return len(dataclasses.asdict(self))

    @abstractmethod
    def __eq__(self, other) -> bool:
        ...

    @abstractmethod
    def __hash__(self) -> int:
        ...

    @abstractmethod
    def __call__(self, backend: Optional[str] = None) -> Any:
        ...

    @property
    @abstractmethod
    def canvas_height(self) -> int:
        ...

    @property
    @abstractmethod
    def canvas_width(self) -> int:
        ...

    @abstractmethod
    def set_canvas_size(self, height: int, width: int) -> 'ElementBuilder':
        ...

    def eval_spec(self, metadata: Mapping[str, str]) -> 'ElementBuilder':
        return self


@dataclasses.dataclass(frozen=True)
class ScalarBarBuilder(ElementBuilder):
    """Addressable container for scalar bar parameters."""
    mapper: Optional[cm.ScalarMappable]
    name: Optional[str] = SCALAR_BAR_DEFAULT_NAME
    name_suffix: Optional[str] = SCALAR_BAR_DEFAULT_NAME_SUFFIX
    below_color: Optional[str] = SCALAR_BAR_DEFAULT_BELOW_COLOR
    length: int = SCALAR_BAR_DEFAULT_LENGTH
    width: int = SCALAR_BAR_DEFAULT_WIDTH
    orientation: Literal['h', 'v'] = SCALAR_BAR_DEFAULT_ORIENTATION
    num_sig_figs: int = SCALAR_BAR_DEFAULT_NUM_SIG_FIGS
    font: str = SCALAR_BAR_DEFAULT_FONT
    name_fontsize_multiplier: float = (
        SCALAR_BAR_DEFAULT_NAME_FONTSIZE_MULTIPLIER
    )
    lim_fontsize_multiplier: float = (
        SCALAR_BAR_DEFAULT_LIM_FONTSIZE_MULTIPLIER
    )
    font_color: Any = SCALAR_BAR_DEFAULT_FONT_COLOR
    font_outline_color: Any = (
        SCALAR_BAR_DEFAULT_FONT_OUTLINE_COLOR
    )
    font_outline_multiplier: float = (
        SCALAR_BAR_DEFAULT_FONT_OUTLINE_MULTIPLIER
    )
    priority: int = 0

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, ScalarBarBuilder):
            return False
        mapper = self.mapper
        if mapper is None or mapper.norm is None:
            return self.name == other.name
        other_mapper = other.mapper
        if other_mapper is None or other_mapper.norm is None:
            return self.name == other.name
        return (
            self.name == other.name
            and self.name_suffix == other.name_suffix
            and np.isclose(mapper.norm.vmin, other.mapper.norm.vmin)
            and np.isclose(mapper.norm.vmax, other_mapper.norm.vmax)
        )

    def __hash__(self) -> int:
        mapper = self.mapper
        if mapper is None or mapper.norm is None:
            return hash((self.name, self.name_suffix, 'ScalarBarBuilder'))
        return hash((
            self.name,
            self.name_suffix,
            mapper.norm.vmin,
            mapper.norm.vmax,
            'ScalarBarBuilder'
        ))

    def __call__(self, backend: Optional[str] = None) -> Any:
        if backend is None:
            backend = get_default_backend()
        return build_scalar_bar(
            **dataclasses.asdict(self),
            backend=backend,
        )

    @property
    def canvas_height(self) -> int:
        if self.orientation == 'h':
            return self.width
        else:
            return self.length

    @property
    def canvas_width(self) -> int:
        if self.orientation == 'h':
            return self.length
        else:
            return self.width

    def set_canvas_size(self, height: int, width: int) -> 'ScalarBarBuilder':
        if self.orientation == 'h':
            return dataclasses.replace(self, width=height, length=width)
        else:
            return dataclasses.replace(self, width=width, length=height)

    def eval_spec(self, metadata: Mapping[str, str]) -> 'ScalarBarBuilder':
        if self.name is None:
            return self
        return dataclasses.replace(self, name=self.name.format(**metadata))


@dataclasses.dataclass(frozen=True)
class TextBuilder(ElementBuilder):
    """Addressable container for text parameters."""
    content: str = TEXT_DEFAULT_CONTENT
    font: str = TEXT_DEFAULT_FONT
    font_size_multiplier: int = TEXT_DEFAULT_FONT_SIZE_MULTIPLIER
    font_color: Any = TEXT_DEFAULT_FONT_COLOR
    font_outline_color: Any = TEXT_DEFAULT_FONT_OUTLINE_COLOR
    font_outline_multiplier: float = TEXT_DEFAULT_FONT_OUTLINE_MULTIPLIER
    bounding_box_width: Optional[int] = TEXT_DEFAULT_BOUNDING_BOX_WIDTH
    bounding_box_height: Optional[int] = TEXT_DEFAULT_BOUNDING_BOX_HEIGHT
    angle: int = TEXT_DEFAULT_ANGLE
    priority: int = 0

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, TextBuilder):
            return False
        return all(
            self[key] == other[key]
            for key in dataclasses.asdict(self)
        )

    def __hash__(self) -> int:
        return hash(
            tuple(dataclasses.asdict(self).values())
        )

    def __call__(self, backend: Optional[str] = None):
        if backend is None:
            backend = get_default_backend()
        return build_text_box(**dataclasses.asdict(self))

    @property
    def canvas_height(self) -> int:
        return self.bounding_box_height

    @property
    def canvas_width(self) -> int:
        return self.bounding_box_width

    def set_canvas_size(self, height, width) -> 'TextBuilder':
        return dataclasses.replace(
            self,
            bounding_box_height=height,
            bounding_box_width=width,
        )

    def eval_spec(self, metadata: Mapping[str, str]) -> 'TextBuilder':
        return dataclasses.replace(
            self,
            content=self.content.format(**metadata)
        )


@dataclasses.dataclass(frozen=True)
class RasterBuilder(ElementBuilder):
    """Addressable container for raster parameters."""
    content: Tensor
    bounding_box_height: int = RASTER_DEFAULT_BOUNDING_BOX_HEIGHT
    bounding_box_width: int = RASTER_DEFAULT_BOUNDING_BOX_WIDTH
    fmt: str = RASTER_DEFAULT_FORMAT
    priority: int = 0

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, RasterBuilder):
            return False
        return np.all(self.content == other.content) and (
            all(
                self[key] == other[key]
                for key in dataclasses.asdict(self)
                if key != 'content'
            )
        )

    def __hash__(self) -> int:
        return hash(
            (self.content.data.tobytes(),) + tuple(
                v for k, v in dataclasses.asdict(self).items()
                if k != 'content'
            )
        )

    def __call__(self, backend: Optional[str] = None):
        if backend is None:
            backend = get_default_backend()
        return build_raster(**dataclasses.asdict(self))

    @property
    def canvas_height(self) -> int:
        return self.height

    @property
    def canvas_width(self) -> int:
        return self.width

    def set_canvas_size(self, height, width) -> 'RasterBuilder':
        return dataclasses.replace(
            self,
            bounding_box_height=height,
            bounding_box_width=width,
        )

    def eval_spec(self, metadata: Mapping[str, str]) -> 'RasterBuilder':
        return self


@dataclasses.dataclass(frozen=True)
class UnknownBuilder(ElementBuilder):
    """Addressable container for arbitrary SVG content."""
    content: str = dataclasses.field(repr=False)
    height: int
    width: int
    priority: int = 0
    _orig_height: Optional[int] = None
    _orig_width: Optional[int] = None

    def __post_init__(self):
        content_sanitised = re.sub('\\<\\?xml.*\n', '', self.content)
        object.__setattr__(
            self,
            'content',
            re.sub('\\<\\!DOCTYPE.*\n', '', content_sanitised)
        )
        if self._orig_height is None:
            object.__setattr__(self, '_orig_height', self.height)
        if self._orig_width is None:
            object.__setattr__(self, '_orig_width', self.width)

    def __eq__(self, other) -> bool:
        if not isinstance(other, RasterBuilder):
            return False
        return all(
            self[key] == other[key]
            for key in dataclasses.asdict(self)
        )

    def __hash__(self) -> int:
        return hash(
            tuple(
                v for k, v in dataclasses.asdict(self).items()
                if k != 'content'
            )
        )

    def __call__(self, backend: Optional[str] = None) -> Any:
        if backend is None:
            backend = get_default_backend()
        return build_unknown(**dataclasses.asdict(self))

    @property
    def canvas_height(self) -> int:
        return self.height

    @property
    def canvas_width(self) -> int:
        return self.width

    def set_canvas_size(self, height: int, width: int) -> 'UnknownBuilder':
        return dataclasses.replace(
            self,
            height=height,
            width=width,
        )

    def eval_spec(self, metadata: Mapping[str, str]) -> 'UnknownBuilder':
        return self


def get_default_backend() -> str:
    return 'svg'


def build_text_box(
    *,
    content: str = TEXT_DEFAULT_CONTENT,
    font: str = TEXT_DEFAULT_FONT,
    font_size_multiplier: int = TEXT_DEFAULT_FONT_SIZE_MULTIPLIER,
    font_color: Any = TEXT_DEFAULT_FONT_COLOR,
    font_outline_color: Any = TEXT_DEFAULT_FONT_OUTLINE_COLOR,
    font_outline_multiplier: float = TEXT_DEFAULT_FONT_OUTLINE_MULTIPLIER,
    bounding_box_width: int = TEXT_DEFAULT_BOUNDING_BOX_WIDTH,
    bounding_box_height: int = TEXT_DEFAULT_BOUNDING_BOX_HEIGHT,
    angle: int = TEXT_DEFAULT_ANGLE,
    backend: Optional[str] = None,
    **params: Any,
):
    if backend is None:
        backend = get_default_backend()
    params = {
        'content': content,
        'font': font,
        'font_size_multiplier': font_size_multiplier,
        'font_color': font_color,
        'font_outline_color': font_outline_color,
        'font_outline_multiplier': font_outline_multiplier,
        'bounding_box_width': bounding_box_width,
        'bounding_box_height': bounding_box_height,
        'angle': angle,
    }
    if backend == 'matplotlib':
        return build_text_box_matplotlib(**params)
    elif backend == 'svg':
        return build_text_box_svg(**params)
    raise ValueError(f'Unknown backend {backend}')


def build_text_box_matplotlib(
    *,
    content: str = TEXT_DEFAULT_CONTENT,
    font: str = TEXT_DEFAULT_FONT,
    font_size_multiplier: int = TEXT_DEFAULT_FONT_SIZE_MULTIPLIER,
    font_color: Any = TEXT_DEFAULT_FONT_COLOR,
    font_outline_color: Any = TEXT_DEFAULT_FONT_OUTLINE_COLOR,
    font_outline_multiplier: float = TEXT_DEFAULT_FONT_OUTLINE_MULTIPLIER,
    bounding_box_width: int = TEXT_DEFAULT_BOUNDING_BOX_WIDTH,
    bounding_box_height: int = TEXT_DEFAULT_BOUNDING_BOX_HEIGHT,
    angle: int = TEXT_DEFAULT_ANGLE,
):
    """
    Build a text box using matplotlib.

    It is strongly advised to use the SVG backend instead.
    """
    figsize = (
        bounding_box_width / TYPICAL_DPI,
        bounding_box_height / TYPICAL_DPI,
    )
    f, ax = plt.subplots(figsize=figsize)
    font_size = 1
    text = ax.text(
        0.5,
        0.5,
        content,
        ha='center',
        va='center',
        fontname=font,
        fontsize=font_size,
        color=font_color,
        rotation=angle,
    )
    # Rescale the text to fit the bounding box
    text_size = text.get_window_extent(f.canvas.get_renderer())
    text_width = text_size.width
    text_height = text_size.height
    scale = min(
        bounding_box_width / text_width,
        bounding_box_height / text_height,
    )
    font_size *= (scale * font_size_multiplier)
    text.set_fontsize(font_size)
    if font_outline_color is not None:
        text.set_path_effects(
            [
                patheffects.withStroke(
                    linewidth=font_size * font_outline_multiplier,
                    foreground=font_outline_color,
                ),
                patheffects.Normal(),
            ]
        )
    ax.axis('off')
    f.subplots_adjust(0, 0, 1, 1)
    return f


def build_text_box_svg(
    *,
    content: str,
    font: str,
    font_size_multiplier: int,
    font_color: Any,
    font_outline_color: Any,
    font_outline_multiplier: float,
    bounding_box_width: int,
    bounding_box_height: int,
    angle: int,
):
    font_size = bounding_box_height * font_size_multiplier
    transform = []
    if angle != 0:
        transform = [
            svg.Rotate(angle, bounding_box_width / 2, bounding_box_height / 2)
        ]
    text_params = dict(
        x=bounding_box_width / 2,
        y=bounding_box_height / 2,
        text=content,
        font_family=font,
        font_size=font_size,
        fill=font_color,
        text_anchor='middle',
        dominant_baseline='middle',
    )
    text_stroke = None
    if font_outline_color is not None:
        text_stroke = svg.Text(
            **text_params,
            stroke=font_outline_color,
            stroke_width=font_size * font_outline_multiplier,
            stroke_linejoin='round',
        )
    text = svg.Text(**text_params)
    elements = [text_stroke, text]
    textgroup = svg.G(
        elements=[e for e in elements if e is not None],
        transform=transform,
    )
    graphics = svg.SVG(
        width=bounding_box_width,
        height=bounding_box_height,
        elements=[textgroup],
    )
    # Rescale the text to fit the bounding box
    # TODO: Well, we tried. This doesn't work. In particular, svgelements is
    #       useless for this purpose. So instead, our prescription for adding
    #       text is to just use a font size that's big enough to fill the
    #       specified bounding box. The figure builder will take care of then
    #       scaling the bounding box and text to the desired size.
    # buffer = io.StringIO()
    # buffer.write(img.__str__())
    # buffer.seek(0)
    # start_w = font_size_multiplier * bounding_box_width / 2
    # start_h = font_size_multiplier * bounding_box_height / 2
    # max_bbox = (
    #     start_w,
    #     start_h,
    #     bounding_box_width - start_w,
    #     bounding_box_height - start_h,
    # )
    # bbox = (float('inf'), float('inf'), float('-inf'), float('-inf'))
    # for e in svgelements.SVG.parse(buffer).elements():
    #     if isinstance(e, svgelements.SVGText):
    #         try:
    #             e_bbox = e.bbox()
    #             bbox = (
    #                 min(bbox[0], e_bbox[0]),
    #                 min(bbox[1], e_bbox[1]),
    #                 max(bbox[2], e_bbox[2]),
    #                 max(bbox[3], e_bbox[3]),
    #             )
    #         except AttributeError:
    #             continue
    # scale = min(
    #     (max_bbox[2] - max_bbox[0]) / (bbox[2] - bbox[0]),
    #     (max_bbox[3] - max_bbox[1]) / (bbox[3] - bbox[1]),
    # )
    # transform += [
    #     svg.Scale(scale, scale),
    # ]
    # textgroup = svg.G(
    #     elements=[e for e in elements if e is not None],
    #     transform=transform,
    # )
    # graphics = svg.SVG(
    #     width=bounding_box_width,
    #     height=bounding_box_height,
    #     elements=[textgroup],
    #     viewBox=f'0 0 {bounding_box_width} {bounding_box_height}',
    # )
    return graphics


def build_scalar_bar(
    *,
    mapper: cm.ScalarMappable,
    name: Optional[str] = SCALAR_BAR_DEFAULT_NAME,
    name_suffix: Optional[str] = SCALAR_BAR_DEFAULT_NAME_SUFFIX,
    below_color: Optional[str] = SCALAR_BAR_DEFAULT_BELOW_COLOR,
    length: int = SCALAR_BAR_DEFAULT_LENGTH,
    width: int = SCALAR_BAR_DEFAULT_WIDTH,
    orientation: Literal['h', 'v'] = SCALAR_BAR_DEFAULT_ORIENTATION,
    num_sig_figs: int = SCALAR_BAR_DEFAULT_NUM_SIG_FIGS,
    font: str = SCALAR_BAR_DEFAULT_FONT,
    name_fontsize_multiplier: float = (
        SCALAR_BAR_DEFAULT_NAME_FONTSIZE_MULTIPLIER
    ),
    lim_fontsize_multiplier: float = (
        SCALAR_BAR_DEFAULT_LIM_FONTSIZE_MULTIPLIER
    ),
    font_color: Any = SCALAR_BAR_DEFAULT_FONT_COLOR,
    font_outline_color: Any = (
        SCALAR_BAR_DEFAULT_FONT_OUTLINE_COLOR
    ),
    font_outline_multiplier: float = (
        SCALAR_BAR_DEFAULT_FONT_OUTLINE_MULTIPLIER
    ),
    backend: Optional[str] = None,
    **params: Any,
) -> Figure:
    """
    Build a scalar bar.

    Parameters
    ----------
    mapper : cm.ScalarMappable
        The mapper to use for the scalar bar.
    name : str, optional
        The name of the scalar bar.
    below_color : str, optional
        The color to use for values below the range.
    length : int, optional
        The length of the scalar bar.
    width : int, optional
        The width of the scalar bar.
    orientation : {'h', 'v'}, optional
        The orientation of the scalar bar.
    num_sig_figs : int, optional
        The number of significant figures to use for the scalar bar.
    font : str, optional
        The font to use for the scalar bar.
    name_fontsize_multiplier : float, optional
        The font size multiplier to use for the name.
    lim_fontsize_multiplier : float, optional
        The font size multiplier to use for the limits.
    font_color : Any, optional
        The font color to use.
    font_outline_color : Any, optional
        The font outline color to use.
    font_outline_multiplier : float, optional
        The font outline multiplier to use.
    backend : str, optional
        The backend to use for the scalar bar.

    Returns
    -------
    Figure
        The figure containing the scalar bar.
    """
    if backend is None:
        backend = get_default_backend()
    params = {
        'mapper': mapper,
        'name': name,
        'name_suffix': name_suffix,
        'below_color': below_color,
        'length': length,
        'width': width,
        'orientation': orientation,
        'num_sig_figs': num_sig_figs,
        'font': font,
        'name_fontsize_multiplier': name_fontsize_multiplier,
        'lim_fontsize_multiplier': lim_fontsize_multiplier,
        'font_color': font_color,
        'font_outline_color': font_outline_color,
        'font_outline_multiplier': font_outline_multiplier,
    }
    if backend == 'matplotlib':
        return build_scalar_bar_matplotlib(**params)
    elif backend == 'svg':
        return build_scalar_bar_svg(**params)
    raise ValueError(f'Unknown backend: {backend}')


def build_scalar_bar_matplotlib(
    *,
    mapper: cm.ScalarMappable,
    name: Optional[str] = SCALAR_BAR_DEFAULT_NAME,
    name_suffix: Optional[str] = SCALAR_BAR_DEFAULT_NAME_SUFFIX,
    below_color: Optional[str] = SCALAR_BAR_DEFAULT_BELOW_COLOR,
    length: int = SCALAR_BAR_DEFAULT_LENGTH,
    width: int = SCALAR_BAR_DEFAULT_WIDTH,
    orientation: Literal['h', 'v'] = SCALAR_BAR_DEFAULT_ORIENTATION,
    num_sig_figs: int = SCALAR_BAR_DEFAULT_NUM_SIG_FIGS,
    font: str = SCALAR_BAR_DEFAULT_FONT,
    name_fontsize_multiplier: float = (
        SCALAR_BAR_DEFAULT_NAME_FONTSIZE_MULTIPLIER
    ),
    lim_fontsize_multiplier: float = (
        SCALAR_BAR_DEFAULT_LIM_FONTSIZE_MULTIPLIER
    ),
    font_color: Any = SCALAR_BAR_DEFAULT_FONT_COLOR,
    font_outline_color: Any = (
        SCALAR_BAR_DEFAULT_FONT_OUTLINE_COLOR
    ),
    font_outline_multiplier: float = (
        SCALAR_BAR_DEFAULT_FONT_OUTLINE_MULTIPLIER
    ),
) -> Figure:
    if name is not None:
        name_suffix = name_suffix or ''
        name = f'{name}{name_suffix}'
        name = name.upper() # TODO: change this! work into style
    vmin, vmax = mapper.get_clim()

    # TODO: Drop this ridiculous hack after we switch to programmatically
    #       creating SVG files instead of using matplotlib
    aspect = length / width
    width = 128
    length = int(width * aspect)

    static_length = num_sig_figs * width // 2
    dynamic_length = length - 2 * static_length
    dynamic = mapper.to_rgba(np.linspace(vmin, vmax, dynamic_length))
    above = np.tile(mapper.to_rgba(vmax), (static_length, 1))
    if below_color is not None:
        if len(below_color) == 4 and below_color[-1] == 0:
            # Not ideal, but looks better than transparent and too many
            # color bars actually end in black
            below_color = '#444444'
        below = np.tile(colors.to_rgba(below_color), (static_length, 1))
    else:
        below = np.tile(mapper.to_rgba(vmin), (static_length, 1))
    rgba = np.stack(width * [np.concatenate([below, dynamic, above])])

    match orientation:
        case 'h':
            figsize = (length / TYPICAL_DPI, width / TYPICAL_DPI)
            vmin_params = {
                'xy': (0, 0),
                'xytext': (0.02, 0.5),
                'rotation': 0,
                'ha': 'left',
                'va': 'center',
            }
            vmax_params = {
                'xy': (1, 0),
                'xytext': (0.98, 0.5),
                'rotation': 0,
                'ha': 'right',
                'va': 'center',
            }
            name_params = {
                'xy': (0.5, 0),
                'xytext': (0.5, 0.5),
                'rotation': 0,
                'ha': 'center',
                'va': 'center',
            }
        case 'v':
            figsize = (width / TYPICAL_DPI, length / TYPICAL_DPI)
            rgba = rgba.swapaxes(0, 1)[::-1]
            vmin_params = {
                'xy': (0, 0),
                'xytext': (0.5, 0.02),
                'rotation': 90,
                'ha': 'center',
                'va': 'bottom',
            }
            vmax_params = {
                'xy': (0, 1),
                'xytext': (0.5, 0.98),
                'rotation': 90,
                'ha': 'center',
                'va': 'top',
            }
            name_params = {
                'xy': (0, .5),
                'xytext': (0.5, 0.5),
                'rotation': 90,
                'ha': 'center',
                'va': 'center',
            }
        case _:
            raise ValueError(f'Invalid orientation: {orientation}')

    f, ax = plt.subplots(figsize=figsize)
    ax.imshow(rgba)
    # TODO: Drop this hack after we switch to programmatically creating
    #       SVG files instead of using matplotlib
    # mul = TYPICAL_DPI / f.dpi
    for vlim, vlim_params in zip(
        (vmin, vmax),
        (vmin_params, vmax_params),
    ):
        fontsize = width * lim_fontsize_multiplier # * mul
        ax.annotate(
            f'{vlim:.{num_sig_figs}g}',
            xycoords='axes fraction',
            fontsize=fontsize,
            fontfamily=font,
            color=font_color,
            path_effects=[
                patheffects.withStroke(
                    linewidth=font_outline_multiplier * fontsize,
                    foreground=font_outline_color,
                )
            ],
            **vlim_params,
        )
    if name is not None:
        fontsize = width * name_fontsize_multiplier # * mul
        ax.annotate(
            name,
            xycoords='axes fraction',
            fontsize=fontsize,
            fontfamily=font,
            color=font_color,
            path_effects=[
                patheffects.withStroke(
                    linewidth=font_outline_multiplier * fontsize,
                    foreground=font_outline_color,
                )
            ],
            **name_params,
        )
    ax.axis('off')
    f.subplots_adjust(0, 0, 1, 1)
    return f


def build_scalar_bar_svg(
    *,
    mapper: cm.ScalarMappable,
    name: Optional[str] = SCALAR_BAR_DEFAULT_NAME,
    name_suffix: Optional[str] = SCALAR_BAR_DEFAULT_NAME_SUFFIX,
    below_color: Optional[str] = SCALAR_BAR_DEFAULT_BELOW_COLOR,
    length: int = SCALAR_BAR_DEFAULT_LENGTH,
    width: int = SCALAR_BAR_DEFAULT_WIDTH,
    orientation: Literal['h', 'v'] = SCALAR_BAR_DEFAULT_ORIENTATION,
    num_sig_figs: int = SCALAR_BAR_DEFAULT_NUM_SIG_FIGS,
    font: str = SCALAR_BAR_DEFAULT_FONT,
    name_fontsize_multiplier: float = (
        SCALAR_BAR_DEFAULT_NAME_FONTSIZE_MULTIPLIER
    ),
    lim_fontsize_multiplier: float = (
        SCALAR_BAR_DEFAULT_LIM_FONTSIZE_MULTIPLIER
    ),
    font_color: Any = SCALAR_BAR_DEFAULT_FONT_COLOR,
    font_outline_color: Any = (
        SCALAR_BAR_DEFAULT_FONT_OUTLINE_COLOR
    ),
    font_outline_multiplier: float = (
        SCALAR_BAR_DEFAULT_FONT_OUTLINE_MULTIPLIER
    ),
) -> Figure:
    if name is not None:
        name_suffix = name_suffix or ''
        name = f'{name}{name_suffix}'
        name = name.upper() # TODO: change this! work into style
    vmin, vmax = mapper.get_clim()

    static_length = num_sig_figs * width // 2
    dynamic_length = length - 2 * static_length
    dynamic = mapper.to_rgba(np.linspace(vmin, vmax, dynamic_length))

    dynamic_grad = svg.LinearGradient(
        id=f'{hash(name)}-gradient',
        elements=[
            svg.Stop(
                offset=i / len(dynamic),
                stop_color=colors.to_hex(dynamic[i]),
            )
            for i in range(len(dynamic))
        ]
    )
    dynamic = svg.Rect(
        x=static_length,
        y=0,
        width=dynamic_length,
        height=width,
        fill=f'url(#{hash(name)}-gradient)',
    )
    above = svg.Rect(
        x=length / 2,
        y=0,
        width=length / 2,
        height=width,
        fill=colors.to_hex(mapper.to_rgba(vmax)),
    )
    if below_color is not None:
        if len(below_color) == 4 and below_color[-1] == 0:
            # Not ideal, but looks better than transparent and too many
            # color bars actually end in black
            below_color = '#444444'
        else:
            below_color = colors.to_hex(below_color)
    else:
        below_color = colors.to_hex(mapper.to_rgba(vmin))
    below = svg.Rect(
        x=0,
        y=0,
        width=length / 2,
        height=width,
        fill=below_color,
    )

    font_color = colors.to_hex(font_color)
    font_outline_color = colors.to_hex(font_outline_color)

    common_params = {
        'y': 0.5 * width,
        'font_family': font,
        'fill': font_color,
        'dominant_baseline': 'middle',
    }
    vlim_params = {
        **common_params,
        'font_size': width * lim_fontsize_multiplier,
    }
    vmin_params = {
        'x': 0.02 * length,
        'text': f'{vmin:.{num_sig_figs}g}',
        'text_anchor': 'start',
        **vlim_params,
    }
    vmax_params = {
        'x': 0.98 * length,
        'text': f'{vmax:.{num_sig_figs}g}',
        'text_anchor': 'end',
        **vlim_params,
    }
    name_params = {
        'x': 0.5 * length,
        'text': name,
        'font_size': width * name_fontsize_multiplier,
        'text_anchor': 'middle',
        **common_params,
    }

    vmin_label_stroke = svg.Text(
        **vmin_params,
        stroke=font_outline_color,
        stroke_width=(
            font_outline_multiplier * width * lim_fontsize_multiplier
        ),
        stroke_linejoin='round',
    )
    vmin_label = svg.Text(**vmin_params)
    vmax_label_stroke = svg.Text(
        **vmax_params,
        stroke=font_outline_color,
        stroke_width=(
            font_outline_multiplier * width * lim_fontsize_multiplier
        ),
        stroke_linejoin='round',
    )
    vmax_label = svg.Text(**vmax_params)
    if name is not None:
        name_label_stroke = svg.Text(
            **name_params,
            stroke=font_outline_color,
            stroke_width=(
                font_outline_multiplier * width * name_fontsize_multiplier
            ),
            stroke_linejoin='round',
        )
        name_label = svg.Text(**name_params)
    else:
        name_label_stroke = None
        name_label = None
    elements = [
        dynamic_grad,
        above,
        below,
        dynamic,
        vmin_label_stroke,
        vmax_label_stroke,
        name_label_stroke,
        vmin_label,
        vmax_label,
        name_label,
    ]
    elements = [e for e in elements if e is not None]
    match orientation:
        case 'h':
            transform =[]
        case 'v':
            transform = [svg.Rotate(-90, 0, 0), svg.Translate(-length, 0)]
            # We nest the elements doubly so that future transforms applied
            # over these will use the canvas coordinate system instead of the
            # rotated coordinate system
            elements = [
                svg.G(
                    elements=elements,
                    transform=transform,
                )
            ]
    group = svg.G(
        id=f'scalarbar-{name}',
        elements=elements,
        transform=[],
    )
    return svg.SVG(
        width=length if orientation == 'h' else width,
        height=width if orientation == 'h' else length,
        elements=[group],
        viewBox=f'0 0 {length} {width}',
    )


def build_raster(
    *,
    content: Tensor,
    bounding_box_height: int = RASTER_DEFAULT_BOUNDING_BOX_HEIGHT,
    bounding_box_width: int = RASTER_DEFAULT_BOUNDING_BOX_WIDTH,
    fmt: str = RASTER_DEFAULT_FORMAT,
    backend: Optional[str] = None,
    **params: Any,
) -> Any:
    if backend is None:
        backend = get_default_backend()
    params = {
        'content': content,
        'bounding_box_height': bounding_box_height,
        'bounding_box_width': bounding_box_width,
        'fmt': fmt,
    }
    if backend == 'svg':
        return build_raster_svg(**params)
    raise ValueError(f'Unknown backend {backend}')


def build_raster_svg(
    *,
    content: Tensor,
    bounding_box_height: int = RASTER_DEFAULT_BOUNDING_BOX_HEIGHT,
    bounding_box_width: int = RASTER_DEFAULT_BOUNDING_BOX_WIDTH,
    fmt: str = RASTER_DEFAULT_FORMAT,
) -> svg.SVG:
    content = Image.fromarray(content)
    if fmt == 'png':
        content = content.convert('RGBA')
    else:
        raise ValueError(
            f'Unsupported format {fmt}. Only png is supported at the '
            'moment.'
        )
    buffer = io.BytesIO()
    content.save(buffer, format=fmt)
    buffer.seek(0)

    image_width, image_height = content.size
    scale = min(
        bounding_box_width / image_width, bounding_box_height / image_height
    )
    image_width = int(image_width * scale)
    image_height = int(image_height * scale)
    shift_x = (bounding_box_width - image_width) // 2
    shift_y = (bounding_box_height - image_height) // 2
    transform = [svg.Translate(shift_x, shift_y)]

    # By convention, we wrap it in a singleton group before returning
    img = svg.Image(
        href='data:image/png;base64,' + base64.b64encode(
            buffer.read()
        ).decode('ascii'),
        width=image_width,
        height=image_height,
    )
    return svg.SVG(
        width=bounding_box_width,
        height=bounding_box_height,
        elements=[svg.G(elements=[img], transform=transform)],
        viewBox=f'0 0 {bounding_box_width} {bounding_box_height}',
    )


@dataclasses.dataclass
class UnknownSVGElement(svg.Element):
    element_name = 'unknown'
    content: str = ''

    def as_str(self) -> str:
        return self.content


def build_unknown(
    *,
    content: str,
    height: int,
    width: int,
    _orig_height: int,
    _orig_width: int,
    backend: Optional[str] = None,
    **params: Any,
):
    if backend is None:
        backend = get_default_backend()
    params = {
        'content': content,
        'height': height,
        'width': width,
        'orig_height': _orig_height,
        'orig_width': _orig_width,
    }
    if backend == 'svg':
        return build_unknown_svg(**params)
    raise ValueError(f'Unknown backend {backend}')


def build_unknown_svg(
    *,
    content: str,
    height: int,
    width: int,
    orig_height: int,
    orig_width: int,
):
    elements = [UnknownSVGElement(content=content)]
    scale_factor = min(height / orig_height, width / orig_width)
    transform = [svg.Scale(scale_factor, scale_factor)]
    #transform = []
    # By convention, we wrap it in a singleton group before returning
    return svg.SVG(
        width=width,
        height=height,
        elements=[
            svg.G(
                elements=[svg.G(elements=elements, transform=transform)],
                transform=[],
            ),
        ],
        viewBox=f'0 0 {width} {height}',
    )


def _uniquify_names(builders: Sequence[ElementBuilder]):
    unique_builders = set()
    retained_builders = []
    for builder in builders:
        if builder not in unique_builders:
            unique_builders.add(builder)
            retained_builders.append(builder)
    return retained_builders


def tile_plot_elements(
    builders: Sequence[ElementBuilder],
    spacing: float = SCALAR_BAR_DEFAULT_SPACING,
    max_dimension: Optional[Tuple[int, int]] = None,
    require_unique_names: bool = True,
) -> Optional[Tensor]:
    builders = [b for b in builders if b is not None]
    if len(builders) == 0:
        return None
    # Algorithm from https://stackoverflow.com/a/28268965
    if require_unique_names:
        builders = _uniquify_names(builders)
    builders = sorted(builders, key=lambda b: b.priority)
    count = len(builders)
    max_width, max_height = max_dimension
    if spacing < 1:
        spacing = spacing * min(max_width, max_height)
    spacing = int(spacing)
    width = median([
        builder.canvas_width
        for builder in builders
    ])
    height = median([
        builder.canvas_height
        for builder in builders
    ])
    candidates = np.concatenate((
        (max_width - np.arange(count) * spacing) /
        (np.arange(1, count + 1) * width),
        (max_height - np.arange(count) * spacing) /
        (np.arange(1, count + 1) * height),
    ))
    candidates = np.sort(candidates[candidates > 0])
    # Vectorising is probably faster than a binary search, so that's what
    # we're doing here
    n_col = np.floor((max_width + spacing) / (width * candidates + spacing))
    n_row = np.floor((max_height + spacing) / (height * candidates + spacing))
    feasible = (n_row * n_col) >= count
    layout_idx = np.max(np.where(feasible))
    layout = np.array([n_row[layout_idx], n_col[layout_idx]], dtype=int)
    scalar = candidates[layout_idx]
    width = int(scalar * width)
    height = int(scalar * height)
    # End algorithm from https://stackoverflow.com/a/28268965

    builders_scaled = []
    for builder in builders:
        builder_width = builder.canvas_width
        builder_height = builder.canvas_height
        scale = min(width / builder_width, height / builder_height)
        builder = builder.set_canvas_size(
            height=int(scale * builder_height),
            width=int(scale * builder_width),
        )
        builders_scaled.append(builder)

    # layout is in array convention, i.e. (nrow, ncol)
    # max_dimension, active_canvas_size, and offset are in Cartesian
    # convention, i.e. (width, height)
    # Things are about to get confusing because we're switching back and forth
    # between these conventions. Sorry.
    #tight = {0: 'row', 1: 'col'}[np.argmin(layout)]
    active_canvas_size_guess = (
        layout[1] * width + (layout[1] - 1) * spacing,
        layout[0] * height + (layout[0] - 1) * spacing,
    )
    argtight = 1 - np.argmin([
        i - j
        for i, j in zip(max_dimension, active_canvas_size_guess)
    ])
    argslack = 1 - argtight
    counttight = layout[argtight]
    countslack = np.ceil(count / counttight).astype(int)
    layout = [None, None]
    layout[argtight] = counttight
    layout[argslack] = countslack

    active_canvas_size = (
        layout[1] * width + (layout[1] - 1) * spacing,
        layout[0] * height + (layout[0] - 1) * spacing,
    )
    offset = [0, 0]
    offset[1 - argslack] = (
        max_dimension[1 - argslack] - active_canvas_size[1 - argslack]
    ) // 2
    base_offset = [o for o in offset]
    canvas_elements = []
    for i, builder in enumerate(builders_scaled):
        element = builder().elements[0]
        internal_displacement = [
            (width - builder.canvas_width) // 2,
            (height - builder.canvas_height) // 2,
        ]
        transform = element.transform + [
            svg.Translate(
                offset[0] + internal_displacement[0],
                offset[1] + internal_displacement[1],
            )
        ]
        element.transform = transform
        canvas_elements.append(element)
        # Filling this in row-major order
        if i % layout[1] == layout[1] - 1:
            offset[0] = base_offset[0]
            offset[1] += height + spacing
        else:
            offset[0] += width + spacing

    collection = svg.G(
        id='elements2d-collection',
        elements=canvas_elements,
    )
    canvas = svg.SVG(
        width=max_width,
        height=max_height,
        viewBox=f'0 0 {max_width} {max_height}',
        elements=[collection],
    )

    return canvas



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from .elements import (
    ElementBuilder,
    RasterBuilder,
    ScalarBarBuilder,
    TextBuilder,
)
from .flows import plotdef, add_network_data
from .layout import (
    AnnotatedLayout,
    Cell,
    CellLayout,
    ColGroupSpec,
    GroupSpec,
    RowGroupSpec,
    grid,
    hsplit,
    vsplit,
)
from .plot import unified_plotter
from .transforms import (
    surf_from_archive,
    surf_from_freesurfer,
    surf_from_gifti,
    surf_scalars_from_cifti,
    surf_scalars_from_gifti,
    surf_scalars_from_array,
    surf_scalars_from_nifti,
    points_scalars_from_nifti,
    points_scalars_from_array,
    parcellate_colormap,
    parcellate_surf_scalars,
    scatter_into_parcels,
    vertex_to_face,
    add_surface_overlay,
    add_points_overlay,
    add_network_overlay,
    build_network,
    node_coor_from_parcels,
    add_node_variable,
    add_edge_variable,
    scalar_focus_camera,
    closest_ortho_camera,
    planar_sweep_camera,
    auto_camera,
    plot_to_image,
    plot_final_image,
    plot_to_html,
    plot_to_display,
    save_snapshots,
    save_figure,
    save_grid,
    svg_element,
    text_element,
    pyplot_element,
)



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Unit tests for elementary brain volume visualisations
"""
from hyve_examples import get_pain_thresh_nifti
from hyve.flows import plotdef
from hyve.transforms import (
    surf_from_archive,
    points_scalars_from_nifti,
    plot_to_image,
    save_snapshots,
    save_grid,
)

import pyvista as pv

def test_vol_scalars():
    nii = get_pain_thresh_nifti()
    plot_f = plotdef(
        surf_from_archive(),
        points_scalars_from_nifti('pain'),
        plot_to_image(),
        save_snapshots(
            fname_spec=(
                'scalars-{pointsscalars}_view-{view}'
            ),
        ),
    )
    plot_f(
        template='fsaverage',
        surf_projection=('pial',),
        surf_alpha=0.3,
        pain_nifti=nii,
        points_scalars_cmap='magma',
        views=('dorsal', 'left', 'anterior'),
        output_dir='/tmp',
    )

def test_vol_scalars_fig():
    nii = get_pain_thresh_nifti()
    plot_f = plotdef(
        surf_from_archive(),
        points_scalars_from_nifti('pain'),
        plot_to_image(),
        save_grid(
            n_cols=3, n_rows=1,
            canvas_size=(1800, 500),
            canvas_color=(1, 1, 1),
            fname_spec=f'scalars-pain_view-all_page-{{page}}',
            scalar_bar_action='collect',
        ),
    )
    plot_f(
        template='fsaverage',
        surf_projection=('pial',),
        surf_alpha=0.3,
        pain_nifti=nii,
        points_scalars_cmap='magma',
        views=('dorsal', 'left', 'anterior'),
        window_size=(1200, 1000),
        output_dir='/tmp',
    )



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Visualisation control flow
~~~~~~~~~~~~~~~~~~~~~~~~~~
Functions for transforming the control flow of visualisation functions.
See also ``transforms.py`` for functions that transform the input and output
flows of visualisation functions.
"""
import inspect
from typing import Any, Literal, Optional, Sequence

from conveyant import emulate_assignment, ichain, join, splice_docstring

from .const import DOCBASE, RETBASE, docbuilder
from .prim import automap_unified_plotter_p


def plotdef(*pparams: Sequence[callable]) -> callable:
    plot_f = ichain(*pparams)(automap_unified_plotter_p)
    # drop variadic parameters
    plot_f.__signature__ = inspect.signature(plot_f).replace(
        parameters=tuple(
            p for p in plot_f.__signature__.parameters.values()
            if p.kind != p.VAR_POSITIONAL
        )
    )
    plot_f = emulate_assignment()(plot_f)
    # TODO: build docstring here when it's done
    return splice_docstring(
        f=plot_f,
        template=docbuilder(),
        base_str=DOCBASE,
        returns=RETBASE,
    )


def _get_unique_parameters_and_make_signature(
    joined: callable,
    fs: Sequence[callable],
) -> callable:
    seen_params = set()
    unique_params = []
    for f in fs:
        f_params = []
        for name, param in f.__signature__.parameters.items():
            if name not in seen_params:
                seen_params.add(name)
                f_params.append(param)
        unique_params = f_params + unique_params
    joined.__signature__ = inspect.signature(joined).replace(
        parameters=unique_params
    )
    return joined


def joindata(
    join_vars: Optional[Sequence[str]] = None,
    how: Literal['outer', 'inner'] = 'outer',
    fill_value: Any = None,
) -> callable:
    def joining_f(arg):
        arg = list(a for a in arg if a is not None)
        out = arg[0].join(arg[1:], how=how)
        if fill_value is not None:
            out = out.fillna(fill_value)
        return out

    return join(
        joining_f,
        join_vars,
        postprocess=_get_unique_parameters_and_make_signature,
    )


def add_network_data(
    *pparams: Sequence[callable],
    how: Literal['outer', 'inner'] = 'outer',
    fill_value: Any = 0.0,
) -> callable:
    return joindata(
        join_vars= ('edge_values', 'node_values'),
        how=how,
        fill_value=fill_value,
    )(*pparams)



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Unit tests using the parcellations in the OHBM ``sparque`` poster
"""
import pytest

from hyve_examples import (
    get_schaefer400_cifti,
    get_myconnectomeWard400_nifti,
    get_mscWard400_nifti,
    get_glasser360_gifti,
    get_gordon333_gifti,
)
from hyve.flows import plotdef
from hyve.transforms import (
    surf_from_archive,
    surf_scalars_from_cifti,
    surf_scalars_from_gifti,
    surf_scalars_from_array,
    surf_scalars_from_nifti,
    parcellate_colormap,
    plot_to_image,
    save_snapshots,
)
from hyve.util import sanitise

@pytest.mark.parametrize('cmap', ['network', 'modal'])
@pytest.mark.parametrize('parcellation_name, parcellation_path', [
    ('Schaefer400', get_schaefer400_cifti()),
    ('MyConnectomeWard400', get_myconnectomeWard400_nifti()),
    ('MSCWard400', get_mscWard400_nifti()),
    ('Glasser360', get_glasser360_gifti()),
    ('Gordon333', get_gordon333_gifti()),
])
def test_sparque(parcellation_name, parcellation_path, cmap):

    # import lytemaps
    # from nilearn import datasets, surface
    # fsaverage = lytemaps.datasets.fetch_fsaverage()
    # print(fsaverage['pial'].L)
    # surf_data_L = surface.vol_to_surf(
    #     parcellation_path,
    #     surf_mesh=str(fsaverage["pial"].L),
    #     #inner_mesh=fsaverage["white_left"],
    #     interpolation="nearest",
    # )
    # surf_data_R = surface.vol_to_surf(
    #     parcellation_path,
    #     surf_mesh=str(fsaverage["pial"].R),
    #     #inner_mesh=fsaverage["white_left"],
    #     interpolation="nearest",
    # )
    # print(surf_data_L.shape, surf_data_R.shape)

    paramstr = sanitise(parcellation_name)
    if isinstance(parcellation_path, dict):
        parcellation_path_L, parcellation_path_R = (
            parcellation_path['left'], parcellation_path['right']
        )
        filearg = {
            f'{paramstr}_gifti_left': parcellation_path_L,
            f'{paramstr}_gifti_right': parcellation_path_R,
        }
        transform = surf_scalars_from_gifti(
            parcellation_name,
            is_masked=False,
            apply_mask=True,
            allow_multihemisphere=False,
        )
    elif parcellation_path.endswith('.nii.gz'):
        filearg = {f'{paramstr}_nifti': parcellation_path}
        transform = surf_scalars_from_nifti(
            parcellation_name,
            template='fsLR',
            method='nearest',
            plot=True,
            threshold=0,
            allow_multihemisphere=False,
        )
    elif parcellation_path.endswith('.nii'):  # Not always, but here yes
        filearg = {f'{paramstr}_cifti': parcellation_path}
        transform = surf_scalars_from_cifti(
            parcellation_name,
            plot=True,
            allow_multihemisphere=False,
        )
        

    plot_f = plotdef(
        surf_from_archive(),
        transform,
        parcellate_colormap(parcellation_name, cmap),
        #vertex_to_face(parcellation_name),
        plot_to_image(),
        save_snapshots(
            fname_spec=(
                'project-sparque_scalars-{surfscalars}_hemisphere-{hemisphere}_view-{view}_'
                f'cmap-{cmap}'
            ),
        ),
    )
    plot_f(
        template='fsLR',
        load_mask=True,
        surf_projection=('veryinflated',),
        hemisphere=['left', 'right'],
        views=[
            'lateral', 'medial', 'ventral', 'dorsal', 'anterior', 'posterior',
        ],
        output_dir='/tmp',
        surf_scalars_boundary_color='black',
        # can't do this because it's autofilled by parcellate_colormap
        # surf_scalars_below_color='#555555',
        surf_scalars_boundary_width=3,
        **filearg,
        # We need a test for array inputs
        # **{f'{parcellation_name}_array_left': surf_data_L},
        # **{f'{parcellation_name}_array_right': surf_data_R},
    )
